# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import yaml

from plumbery.polisher import PlumberyPolisher
from plumbery.plogging import plogging


class InventoryPolisher(PlumberyPolisher):
    """
    Captures inventory information

    This polisher looks at each node in sequence, but only to retrieve
    maximum information about each of them. At the end of the process,
    the polisher writes a YAML file for future reference.

    To activate this polisher you have to mention it in the fittings plan,
    like in the following example::

        ---
        actions:
          - inventory:
              output: inventory.yaml
        ---
        # Frankfurt in Europe
        locationId: EU6
        regionId: dd-eu
        ...


    """

    def go(self, engine):
        """
        Restarts the inventory process
        """

        super(InventoryPolisher, self).go(engine)

        self.inventory = []

    def shine_node(self, node, settings, container):
        """
        Gets as much information as possible from a node

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param settings: the fittings plan for this node
        :type settings: ``dict``

        :param container: the container of this node
        :type container: :class:`plumbery.PlumberyInfrastructure`

        Please note that the information saved here is a combination of
        attributes exposed by Apache Libcloud and of extra fields
        provided by Dimension Data.

        """

        plogging.info("Examinating node '{}'".format(settings['name']))
        if node is None:
            plogging.info("- not found")
            return

        data = {}
        data['type'] = 'node'
        data['id'] = node.id
        data['name'] = node.name

        data['state'] = node.state

        status = None
        if 'status' in node.extra:
            status = node.extra.pop('status')
            data['action'] = str(status.action)

        if 'cpu' in node.extra:
            cpu = node.extra.pop('cpu')
            data['cpu'] = cpu.cpu_count
            data['cpuCoresPerSocket'] = cpu.cores_per_socket
            data['cpuPerformance'] = cpu.performance.lower()

        if 'memoryMb' in node.extra:
            memory = node.extra.pop('memoryMb')
            data['memory'] = int(memory/1024)

        if 'networkId' in node.extra:
            node.extra.pop('networkId')

        data.update(node.extra)

        domain = container.get_network_domain(
            container.blueprint['domain']['name'])

        if domain is not None:
            data['datacenter'] = domain.location.name
            data['datacenterCountry'] = domain.location.country
            data['networkDomain'] = domain.name
            data['networkDomainDescription'] = domain.description

        ethernet = container.get_ethernet(
            container.blueprint['ethernet']['name'])

        if ethernet is not None:
            data['ethernet'] = ethernet.name
            data['ethernetId'] = ethernet.id
            data['ethernetDescription'] = ethernet.description

        data['public_ips'] = node.public_ips
        data['private_ips'] = node.private_ips

        data['private_host'] = node.private_ips[0].replace('.', '-')

        tags = []
        tags.append("state_{}".format(data['state']))
        tags.append("location_{}".format(data['datacenterId']))
        if 'networkDomain' in data:
            tags.append("domain_{}".format(data['networkDomain']))
        if 'ethernet' in data:
            tags.append("network_{}".format(data['ethernet']))
        description = node.extra['description'].replace(
            '#plumbery', '').strip()
        tags += {tag.strip("#") for tag in description.split()
                 if tag.startswith("#")}
        data['tags'] = tags

        self.inventory.append(data)

        plogging.info("- done")

    def reap(self):
        """
        Saves information gathered through the polishing sequence

        All information captured in dumped in a file, in YAML format,
        to provide a flexible and accurate inventory of all live nodes
        described in the fittings plan.

        """

        if len(self.inventory) < 1:
            return

        if 'output' in self.settings:
            fileName = self.settings['output']
            plogging.info("Writing inventory in '{}'".format(fileName))
            stream = open(fileName, 'w')
        else:
            plogging.info("Showing the inventory")
            stream = sys.stdout

        stream.write(yaml.dump(self.inventory, default_flow_style=False))



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import six
import string
import yaml
from six import string_types
from plumbery.plogging import plogging
from plumbery.nodes import PlumberyNodes

__all__ = ['PlumberyText', 'PlumberyContext']

if six.PY2:
    b = bytes = ensure_string = str
else:
    def ensure_string(s):
        if isinstance(s, str):
            return s
        elif isinstance(s, bytes):
            return s.decode('utf-8')
        else:
            raise TypeError("Invalid argument %r for ensure_string()" % (s,))


class PlumberyText:

    @classmethod
    def expand_parameters(cls, text, context):
        """
        Binds parameters and produces a string

        :param text: the text to be expanded
        :type text: ``str``

        :param context: context for lookup of parameters
        :type context: :class:`PlumberyContext`

        :return: the expanded text
        :rtype: ``str``

        This function provides low-level binding of parameters.

        """

        opening = '{{'
        closing = '}}'

        if not isinstance(text, string_types):
            raise TypeError("Parameters expansion requires textual input")

        expanded = ''
        index = 0
        debugged = []

        while index < len(text):
            head = text.find(opening, index)
            if head < 0:
                expanded += text[index:]
                break

            tail = text.find(closing, head+len(opening))
            if tail < 0:
                expanded += text[index:]
                break

            while True:
                head_next = text.find(opening, head+len(opening), tail)
                if head_next > head:
                    head = head_next
                else:
                    break

            token = text[head+len(opening):tail].strip(' \\\t')
            if len(token) < 1:
                expanded += text[index:tail+len(closing)]
                index = tail+len(closing)
                continue

            if not token.startswith('parameter.'):
                expanded += text[index:tail+len(closing)]
                index = tail+len(closing)
                continue

            replacement = context.lookup(token)
            if replacement is None:
                raise KeyError("Missing parameter '{}'".format(token))

            if token not in debugged:
                plogging.debug("- '{}' -> '{}'".format(token, replacement))
                debugged.append(token)

            expanded += text[index:head]+str(replacement)
            index = tail+len(closing)

        return expanded

    @classmethod
    def expand_string(cls, text, context):
        """
        Binds variables and produces a string

        :param text: the text or the structure to be expanded
        :type text: ``str`` or ``dict``

        :param context: context for lookup of tokens
        :type context: :class:`PlumberyContext`

        :return: the expanded text
        :rtype: ``str``

        This function allows for dynamic binding of data known by plumbery.

        """

        opening = '{{'
        closing = '}}'

        serialized = False
        if not isinstance(text, string_types):  # serialize python object
            plogging.debug("- serializing object before expansion")
            if six.PY2:
                text = str(text)
            else:
                text = str(text)
            serialized = True

        expanded = ''
        index = 0
        debugged = []

        while index < len(text):
            head = text.find(opening, index)
            if head < 0:
                expanded += text[index:]
                break

            tail = text.find(closing, head+len(opening))
            if tail < 0:
                expanded += text[index:]
                break

            token = text[head+len(opening):tail].strip(' \\\t')
            if len(token) < 1:
                expanded += text[index:tail+len(closing)]
                index = tail+len(closing)
                continue

            replacement = context.lookup(token)
            if replacement is None:   # preserve unmatched tag
                if token not in debugged:
                    plogging.debug("- no match for '{}'".format(token))
                    debugged.append(token)

                expanded += text[index:tail+len(closing)]
                index = tail+len(closing)

            else:  # actual expansion
                if token not in debugged:
                    plogging.debug("- '{}' -> '{}'".format(token, replacement))
                    debugged.append(token)

                if serialized:  # preserve line breaks
                    replacement = ensure_string(replacement).replace('\n', '\\'+'n')

                expanded += text[index:head]+str(replacement)
                index = tail+len(closing)

        if serialized:  # from serialized python to yaml representation

            # protect  \ followed by \
            watermark1 = '-=_+*=-'
            expanded = expanded.replace('\\'+'\\', watermark1+'|'+watermark1)

            # protect  \ followed by '
            watermark2 = '-=*+_=-'
            expanded = expanded.replace('\\'+"'", watermark2+'|'+watermark2)

            # protect  None
            watermark3 = '=-_+*-='
            expanded = expanded.replace('None', watermark3+'None'+watermark3)

            instanciated = yaml.load(expanded)
            expanded = PlumberyText.dump(instanciated)

            expanded = expanded.replace(watermark1+'|'+watermark1, '\\')
            expanded = expanded.replace(watermark2+'|'+watermark2, "'")
            expanded = expanded.replace(watermark3+'None'+watermark3, 'None')

        return expanded

    @classmethod
    def could_expand(cls, content):
        """
        Checks if some bytes are expandable or not

        :param content: the blob to be tested
        :type content: ``str``

        :return: True if this could be expandable, False otherwise
        :rtype: ``bool``

        """

        if str(content[0]) in string.ascii_letters+string.digits+' \t\r\n#/-,.;=':
            return True

        return False

    @classmethod
    def dump(cls, content):
        """
        Transforms a dictionary to a string

        :param content: the dictionary to transform
        :type content: ``dict``

        :return: the related string

        """
        return cls.dump_dict(content, spaces=0).strip()+'\n'

    @classmethod
    def dump_dict(cls, content, spaces=0):
        """
        Transforms a sub-dictionary to a string

        :param content: the dictionary to transform
        :type content: ``dict``

        :param spaces: shift to the right from containing elements
        :type spaces: ``int``

        :return: the related string

        """
        if isinstance(content, int):
            return cls.dump_str(str(content), spaces+2)

        if isinstance(content, str):
            content = yaml.load(content)

        text = ''
        for key in content.keys():
            text += '\n' + ' ' * spaces + key + ': '

            value = content[key]

            if isinstance(value, dict):
                text += cls.dump_dict(value, spaces+2)

            elif isinstance(value, (list, tuple)):
                text += cls.dump_list(value, spaces+2)

            elif isinstance(value, bool):
                text += str(value).lower()

            else:
                text += cls.dump_str(str(value), spaces+2)

        return text

    @classmethod
    def dump_list(cls, content, spaces=0):
        """
        Transforms a list to a string

        :param content: the list to transform
        :type content: ``list``

        :param spaces: shift to the right from containing elements
        :type spaces: ``int``

        :return: the related string

        """
        if isinstance(content, str):
            content = yaml.load(content)

        text = ''
        for item in content:
            text += '\n' + ' ' * spaces + '- '

            if isinstance(item, dict):
                text += cls.dump_dict(item, spaces+2).lstrip()

            elif isinstance(item, list):
                text += cls.dump_list(item, spaces+2)

            elif isinstance(item, tuple):
                text += cls.dump_tuple(item, spaces+2)

            elif isinstance(item, bool):
                text += str(item).lower()

            else:
                text += cls.dump_str(str(item), spaces+2)

        return text

    @classmethod
    def dump_str(cls, content, spaces=0):
        """
        Transforms a string to a string

        :param content: the string to transform
        :type content: ``str``

        :param spaces: shift to the right from containing elements
        :type spaces: ``int``

        :return: the related string

        """

        lines = content.split('\n')
        if len(lines) == 1:              # that's a real hack...
            lines = content.split('\\n')

        if len(lines) == 1:              # quote string if it would fool yaml
            if len(content) > 0 and content[-1] in ('-', '\\', '|'):
                return '"'+content+'"'
            return content

        text = '|'
        spaces += 2
        for line in lines:
            text += '\n'
            if len(line) > 0:
                text += ' ' * spaces + line

        return text


class PlumberyContext:

    def __init__(self, dictionary=None, context=None):

        if dictionary is None or not isinstance(dictionary, dict):
            self.dictionary = {}
        else:
            self.dictionary = dictionary

        self.keys = self.dictionary.keys()

        self.context = context

    def lookup(self, token):
        """
        Retrieves the value attached to a token

        :param token: the token
        :type token: ``str``

        :return: the value attached to this token, or `None`

        """

        if token in self.keys:
            return str(self.dictionary[token])

        if self.context is not None:
            return self.context.lookup(token)

        return None


class PlumberyNodeContext:

    def __init__(self, node, container=None, context=None):

        self.node = node
        self.container = container
        self.context = context
        self.cache = {}

        if node is None:
            return

        self.cache = {
            'node': node.private_ips[0],
            node.name: node.private_ips[0],
            'node.private': node.private_ips[0],
            node.name+'.private': node.private_ips[0],
            'node.ipv6': node.extra['ipv6'],
            node.name+'.ipv6': node.extra['ipv6'],
            'node.id': node.id,
            node.name+'.id': node.id,
            'node.name': node.name,
            node.name+'.name': node.name,
            'node.private_host': node.private_ips[0].replace('.', '-'),
            node.name+'.private_host': node.private_ips[0].replace('.', '-')
            }

        if len(node.public_ips) > 0:
            self.cache['node.public'] = node.public_ips[0]
            self.cache[node.name+'.public'] = node.public_ips[0]

    def lookup(self, token):
        """
        Retrieves the value attached to a token

        :param token: the token, e.g., 'node.ipv6'
        :type token: ``str``

        :return: the value attached to this token, or `None`

        """

        if token in self.cache:
            return str(self.cache[token])

        value = None
        if self.context is not None:
            value = self.context.lookup(token)

        if value is not None:
            return value

        if self.container is None:
            return None

        tokens = token.split('.')
        if len(tokens) < 2:
            tokens.append('private')

        nodes = PlumberyNodes(self.container.facility)
        node = nodes.get_node(tokens[0])
        if node is None:
            return None

        if self.context is not None:
            self.context.remember(tokens[0], node.private_ips[0])
            self.context.remember(tokens[0]+'.private', node.private_ips[0])
            self.context.remember(tokens[0]+'.ipv6', node.extra['ipv6'])
            if len(node.public_ips) > 0:
                self.context.remember(tokens[0]+'.public', node.public_ips[0])

        if tokens[1] == 'private':
            return node.private_ips[0]

        if tokens[1] == 'ipv6':
            return node.extra['ipv6']

        if tokens[1] == 'public':
            if len(node.public_ips) > 0:
                return node.public_ips[0]
            else:
                return ''

        return None



#!/usr/bin/env python

"""
Tests for `action` module.
"""

import unittest

from plumbery.action import PlumberyAction, PlumberyActionLoader
from plumbery.util import PlumberyParameters


class TestPlumberyAction(unittest.TestCase):

    def test_signatures(self):

        class DummyAction(PlumberyAction):
            def __init__(self, settings):
                self.count = 0

            def begin(self, engine):
                self.count += 100

            def enter(self, facility):
                self.count *= 2

            def process(self, blueprint):
                self.count += 1

            def quit(self):
                self.count += 3

            def end(self):
                self.count += 1

        action = DummyAction({})
        action.begin(engine=None)
        action.enter(facility=None)
        action.process(blueprint=None)
        action.process(blueprint=None)
        action.quit()
        action.enter(facility=None)
        action.process(blueprint=None)
        action.process(blueprint=None)
        action.quit()
        action.enter(facility=None)
        action.process(blueprint=None)
        action.process(blueprint=None)
        action.quit()
        action.end()
        self.assertEqual(action.count, 836)

    def test_loader_ok(self):

        action = PlumberyActionLoader.load('noop')
        action.begin(engine=None)
        action.enter(facility=None)
        action.process(blueprint=None)
        action.process(blueprint=None)
        action.quit()
        action.enter(facility=None)
        action.process(blueprint=None)
        action.process(blueprint=None)
        action.quit()
        action.enter(facility=None)
        action.process(blueprint=None)
        action.process(blueprint=None)
        action.quit()
        action.end()

    def test_loader_unknown_class(self):

        with self.assertRaises(ImportError):
            PlumberyActionLoader.load('*123*')

    def test_getters(self):
        action = PlumberyActionLoader.load('noop', {'parameter': 'ok'})
        self.assertEqual(action.get_label(), 'noop')
        self.assertEqual(action.get_parameter('parameter'), 'ok')

        parameters = PlumberyParameters({'goofy': 'ok'})
        action = PlumberyActionLoader.load('noop', parameters)
        self.assertEqual(action.get_label(), 'noop')
        self.assertEqual(action.get_parameter('goofy'), 'ok')

    def test_static_load(self):

        actions = ('ansible',
                   'build',
                   'configure',
                   'destroy',
                   'information',
                   'noop',
                   'ping',
                   'prepare',
                   'start',
                   'stop',
                   'wipe',
                  )

        for label in actions:
            action = PlumberyActionLoader.load(label, {})
            self.assertEqual(action.get_label(), label)
            action.begin(engine=None)
            action.enter(facility=None)
            action.process(blueprint=None)
            action.process(blueprint=None)
            action.quit()
            action.enter(facility=None)
            action.process(blueprint=None)
            action.process(blueprint=None)
            action.quit()
            action.enter(facility=None)
            action.process(blueprint=None)
            action.process(blueprint=None)
            action.quit()
            action.end()

    def test_dynamic_load(self):

        actions = PlumberyActionLoader.load_all()

        expected = ['ansible',
                   'build',
                   'configure',
                   'destroy',
                   'information',
                   'inventory',
                   'noop',
                   'ping',
                   'prepare',
                   'start',
                   'stop',
                   'wipe',
                  ]

        self.assertEqual(sorted(actions.keys()), expected)


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import argparse
import os
import requests
import six
import sys

if six.PY2:
    from urlparse import urljoin as up
else:
    from urllib.parse import urljoin as up


def parse_args(args):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-o', '--output', help='output directory')
    parser.add_argument('files', nargs='+', help='The URL for remote files')
    return parser.parse_args(args)


def main(args):
    if args.output is None:
        args.output = os.getcwd()
    for url in args.files:
        filename = url.split('/')[-1]
        if filename == 'manifest.mf':
            download_manifest(url, args.output)
        else:
            download_file(url, args.output)


def download_manifest(url, output_dir):
    download_file(url, output_dir)
    with open(os.path.join(output_dir, 'manifest.mf'), 'r') as manifest_file:
        lines = manifest_file.readlines()
        for i in lines:
            download_file(up(url, i), output_dir)


def download_file(url, output_dir):
    local_filename = url.split('/')[-1]
    r = requests.get(url, stream=True)
    with open(os.path.join(output_dir, local_filename), 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024): 
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)
    return local_filename

if __name__ == "__main__":
    args = parse_args(sys.argv[1:])
    main(args)


# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from plumbery.polishers.inventory import InventoryPolisher
from plumbery.plogging import plogging


class AnsiblePolisher(InventoryPolisher):
    """
    Captures inventory information for ansible

    This polisher looks at each node in sequence, but only to retrieve
    maximum information about each of them. At the end of the process,
    the polisher writes a YAML file to feed ansible.

    To activate this polisher you have to mention it in the fittings plan,
    like in the following example::

        ---
        safeMode: False
        actions:
          - ansible:
              output: inventory.yaml
        ---
        # Frankfurt in Europe
        locationId: EU6
        regionId: dd-eu
        ...


    """

    def reap(self):
        """
        Saves information gathered through the polishing sequence

        All information captured in dumped in a file, in YAML format,
        to provide a flexible and accurate inventory of live nodes
        that are part of the fittings plan.

        The private ipv4 network addresses are attached to each host, so that
        ansible do not need to rely on DNS settings.

        Plumbery also provides multiple groups of hosts to ansible, based on
        various criteria:

        * data center location
        * hashtags in node description

        As an example, let consider the following fittings plan::

            ---
            locationId: EU6
            regionId: dd-eu
            blueprints:
              - sql:
                  domain: *vdc1
                  ethernet: *data
                  nodes:
                    - slaveSQL:
                        description: 'SQL #vdc1 #secondary'
                        appliance: 'RedHat 6 64-bit 4 CPU'
            ---
            locationId: EU7
            regionId: dd-eu
            blueprints:
              - sql:
                  domain: *vdc2
                  ethernet: *data
                  nodes:
                    - masterSQL:
                        description: 'SQL #vdc2 #primary'
                        appliance: 'RedHat 6 64-bit 4 CPU'


        After the creation of nodes by plumbery, the inventory provided to
        ansible could look like this::

            masterSQL  private_ipv4=10.2.3.8 ansible_ssh_host=10.2.3.8
            slaveSQL  private_ipv4=10.1.3.8 ansible_ssh_ host=10.1.3.8

            [EU6]
            slaveSQL

            [EU7]
            masterSQL

            [plumbery]
            masterSQL
            slaveSQL

            [primary]
            masterSQL

            [secondary]
            slaveSQL

            [vdc1]
            slaveSQL

            [vdc2]
            masterSQL


        Ok, with two nodes it may be quite difficult to find excitment in
        a sophisticated inventory, but you get the idea.

        """

        hosts = []
        groups = {}

        for item in self.inventory:
            host = item['name']

            if len(item['public_ips']) > 0:
                host_address = item['public_ips'][0]
            elif item['ipv6']:
                host_address = item['ipv6']
            else:
                host_address = item['private_ips'][0]

            hosts.append("{} private_ipv4={} ansible_ssh_host={}".format(
                host, host_address, host_address))

            group = 'dc='+item['datacenterId']
            if group not in groups.keys():
                groups[group] = []
            groups[group].append(host)

            description = item['description']
            tags = {tag.strip("#") for tag in description.split()
                    if tag.startswith("#")}
            for tag in tags:
                tag = 'role='+tag
                if tag not in groups.keys():
                    groups[tag] = []
                groups[tag].append(host)

        if 'output' in self.settings:
            fileName = self.settings['output']
            plogging.info("Writing inventory for ansible in '{}'".format(
                fileName))
            stream = open(fileName, 'w')
        else:
            plogging.info("Showing the inventory for ansible")
            stream = sys.stdout

        for line in sorted(hosts):
            stream.write(line+'\n')
        stream.write('\n')

        for group in sorted(groups.keys()):
            stream.write('[{}]\n'.format(group))
            for host in sorted(groups[group]):
                stream.write('{}\n'.format(host))
            stream.write('\n')
        stream.write('\n')



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import re
import time
from socket import error as SocketError
import errno

from libcloud.compute.base import NodeAuthPassword
from libcloud.compute.base import NodeState
from libcloud.utils.xml import fixxpath, findall
from libcloud.common.dimensiondata import TYPES_URN
from libcloud.common.dimensiondata import DimensionDataServerCpuSpecification

from plumbery.exception import PlumberyException
from plumbery.infrastructure import PlumberyInfrastructure
from plumbery.plogging import plogging
from plumbery.util import retry
from plumbery.polishers.monitoring import MonitoringConfiguration

__all__ = ['PlumberyNodes']


class PlumberyNodes(object):
    """
    Cloud automation for computing nodes

    :param facility: the underlying physical facility
    :type facility: :class:`plumbery.PlumberFacility`

    A node is a virtual machine with some permanent storage, and one or several
    network connections. It can have many attributes, such as the number of
    CPU, some memory, an operating system, etc.

    Example::

        from plumbery.nodes import PlumberyNodes
        nodes = PlumberyNodes(facility)
        nodes.build_blueprint(blueprint, container)

    In this example an instance is initialised at the given facility, and then
    it is asked to create nodes described in the provided blueprint.
    This is not covering network and the security, but only the nodes.

    Attributes:
        facility (PlumberyFacility):
            a handle to the physical facility where nodes are deployed
            are implemented

    """

    plumbery = None

    def __init__(self, facility=None):
        """Put nodes in context"""

        # handle to parent parameters and functions
        self.facility = facility
        self.region = facility.region
        self.backup = facility.backup
        self.plumbery = facility.plumbery

    def __repr__(self):

        return "<PlumberyNodes facility: {}>".format(self.facility)

    def build_blueprint(self, blueprint, container):
        """
        Create missing nodes

        :param blueprint: the blueprint to build
        :type blueprint: ``dict``

        :param container: the container where nodes will be built
        :type container: :class:`plumbery.PlumberyInfrastructure`

        """

        plogging.debug("Building nodes of blueprint '{}'".format(
            blueprint['target']))

        self.facility.power_on()

        if ('nodes' not in blueprint
                or not isinstance(blueprint['nodes'], list)):

            plogging.debug("No nodes have been defined in '{}'".format(
                blueprint['target']))
            blueprint['nodes'] = []

        for item in blueprint['nodes']:

            if type(item) is dict:
                label = list(item.keys())[0]
                settings = list(item.values())[0]

            else:
                label = item
                settings = {}

            for label in self.expand_labels(label):

                plogging.info("Creating node '{}'".format(label))

                if self.get_node(label):
                    plogging.info("- already there")
                    continue

                description = '#plumbery'
                if 'description' in settings:
                    description = settings['description'] + ' #plumbery'

                if 'appliance' in settings:
                    imageName = settings['appliance']
                else:
                    imageName = None

                image = self.facility.get_image(imageName)
                if image is None:
                    raise PlumberyException("Error: unable to find image "
                                            "for '{}'!".format(imageName))
                plogging.debug("- using image '{}'".format(image.name))

                cpu = None
                if 'cpu' in settings:
                    tokens = str(settings['cpu']).split(' ')
                    if len(tokens) < 2:
                        tokens.append('1')
                    if len(tokens) < 3:
                        tokens.append('standard')

                    if (int(tokens[0]) < 1
                            or int(tokens[0]) > 32):

                        plogging.info("- cpu should be between 1 and 32")

                    elif (int(tokens[1]) < 1
                            or int(tokens[1]) > 2):

                        plogging.info("- core per cpu should be either 1 or 2")

                    elif tokens[2].upper() not in ('STANDARD',
                                                   'HIGHPERFORMANCE'):

                        plogging.info("- cpu speed should be either 'standard'"
                                     " or 'highspeed'")

                    else:
                        cpu = DimensionDataServerCpuSpecification(
                            cpu_count=tokens[0],
                            cores_per_socket=tokens[1],
                            performance=tokens[2].upper())
                        plogging.debug("- assigning {} cpus".format(
                            cpu.cpu_count))
                        plogging.debug("- core per cpu: {}".format(
                            cpu.cores_per_socket))
                        plogging.debug("- cpu performance: {}".format(
                            cpu.performance.lower()))

                memory = None
                if 'memory' in settings:
                    memory = int(settings['memory'])
                    if memory < 1 or memory > 256:
                        plogging.info("- memory should be between 1 and 256")
                        memory = None
                    else:
                        plogging.debug("- assigning {} GB of memory".format(
                            memory))

                if self.plumbery.safeMode:
                    plogging.info("- skipped - safe mode")
                    continue

                if container.domain is None:
                    plogging.info("- missing network domain")
                    continue

                if container.network is None:
                    plogging.info("- missing Ethernet network")
                    continue

                primary_ipv4 = None
                if 'glue' in settings:
                    for line in settings['glue']:

                        tokens = line.strip(' ').split(' ')
                        token = tokens.pop(0)

                        if token.lower() == 'primary':
                            token = container.network.name

                        if token != container.network.name:
                            continue

                        if len(tokens) < 1:
                            break

                        plogging.info("Glueing node '{}' to network '{}'"
                                     .format(label, token))

                        numbers = tokens.pop(0).strip('.').split('.')
                        subnet = container.network.private_ipv4_range_address.split('.')
                        while len(numbers) < 4:
                            numbers.insert(0, subnet[3-len(numbers)])

                        primary_ipv4 = '.'.join(numbers)
                        plogging.debug("- using address '{}'"
                                      .format(primary_ipv4))

                        break

                retries = 2
                should_start = False
                while True:

                    try:
                        if primary_ipv4 is not None:
                            self.region.create_node(
                                name=label,
                                image=image,
                                auth=NodeAuthPassword(
                                    self.plumbery.get_shared_secret()),
                                ex_network_domain=container.domain,
                                ex_primary_ipv4=primary_ipv4,
                                ex_cpu_specification=cpu,
                                ex_memory_gb=memory,
                                ex_is_started=should_start,
                                ex_description=description)

                        else:
                            self.region.create_node(
                                name=label,
                                image=image,
                                auth=NodeAuthPassword(
                                    self.plumbery.get_shared_secret()),
                                ex_network_domain=container.domain,
                                ex_vlan=container.network,
                                ex_cpu_specification=cpu,
                                ex_memory_gb=memory,
                                ex_is_started=should_start,
                                ex_description=description)

                        plogging.info("- in progress")

                        if should_start:  # stop the node after start

                            plogging.info("- waiting for node to be deployed")
                            node = None
                            while True:
                                node = self.get_node(label)
                                if node is None:
                                    plogging.error("- aborted - missing node '{}'".format(label))
                                    return

                                if node.extra['status'].action is None:
                                    break

                                if (node is not None
                                        and node.extra['status'].failure_reason is not None):

                                    plogging.error("- aborted - failed deployment "
                                                 "of node '{}'".format(label))
                                    return

                                time.sleep(20)

                            if node is not None:
                                self.region.ex_shutdown_graceful(node)
                                plogging.info("- shutting down after deployment")

                    except SocketError as feedback:

                        if feedback.errno == errno.ECONNRESET and retries > 0:
                            retries -= 1
                            time.sleep(10)
                            continue

                        else:
                            plogging.info("- unable to create node")
                            plogging.error(str(feedback))

                    except Exception as feedback:

                        if 'RESOURCE_BUSY' in str(feedback):
                            time.sleep(10)
                            continue

                        elif 'RESOURCE_NOT_FOUND' in str(feedback):
                            plogging.info("- not now")
                            plogging.error(str(feedback))

                        elif 'RESOURCE_LOCKED' in str(feedback):
                            plogging.info("- not now - locked")
                            plogging.error(str(feedback))

                        elif ('INVALID_INPUT_DATA: Cannot deploy server '
                              'with Software Labels in the "Stopped" state.' in
                              str(feedback)):
                            should_start = True
                            continue

                        else:
                            plogging.info("- unable to create node")
                            plogging.error(str(feedback))

                    break

    def destroy_blueprint(self, blueprint):
        """
        Destroys nodes of a given blueprint

        :param blueprint: the blueprint to build
        :type blueprint: ``dict``

        """

        self.facility.power_on()

        infrastructure = PlumberyInfrastructure(self.facility)
        container = infrastructure.get_container(blueprint)

        if ('nodes' not in blueprint
                or not isinstance(blueprint['nodes'], list)):
            return

        # destroy in reverse order
        for item in reversed(blueprint['nodes']):

            if type(item) is dict:
                label = list(item)[0]
                settings = item[label]
            else:
                label = str(item)
                settings = {}

            for label in self.expand_labels(label):

                node = self.get_node(label)
                if node is None:
                    plogging.info("Destroying node '{}'".format(label))
                    plogging.info("- not found")
                    continue

                if 'destroy' in settings and settings['destroy'] == 'never':
                    plogging.info("Destroying node '{}'".format(label))
                    plogging.info("- this node can never be destroyed")
                    return False

                timeout = 300
                tick = 6
                while node.extra['status'].action == 'SHUTDOWN_SERVER':
                    time.sleep(tick)
                    node = self.get_node(label)
                    timeout -= tick
                    if timeout < 0:
                        break

                if node.state == NodeState.RUNNING:
                    plogging.info("Destroying node '{}'".format(label))
                    plogging.info("- skipped - node is up and running")
                    continue

                if self.plumbery.safeMode:
                    plogging.info("Destroying node '{}'".format(label))
                    plogging.info("- skipped - safe mode")
                    continue

                configuration = MonitoringConfiguration(
                    engine=container.facility.plumbery,
                    facility=container.facility)
                configuration.deconfigure(node, settings)

                self._detach_node(node, settings)
                container._detach_node_from_internet(node)

                plogging.info("Destroying node '{}'".format(label))
                while True:

                    try:
                        self.region.destroy_node(node)
                        plogging.info("- in progress")

                    except Exception as feedback:

                        if 'RESOURCE_BUSY' in str(feedback):
                            time.sleep(10)
                            continue

                        elif 'RESOURCE_NOT_FOUND' in str(feedback):
                            plogging.info("- not found")

                        elif 'SERVER_STARTED' in str(feedback):
                            plogging.info("- skipped - node is up and running")

                        elif 'RESOURCE_LOCKED' in str(feedback):
                            plogging.info("- not now - locked")
                            return False

                        else:
                            plogging.info("- unable to destroy node")
                            plogging.error(str(feedback))

                    break

    def _detach_node(self, node, settings):
        """
        Detach a node from multiple networks

        :param node: the target node
        :type node: :class:`libcloud.compute.base.Node`

        This function removes all secondary network interfaces to a node, and
        any potential translation to the public Internet.

        """

        if node is None:
            return True

        if ('running' in settings
                and settings['running'] == 'always'
                and node.state == NodeState.RUNNING):

            return True

        for interface in self._list_secondary_interfaces(node):

            plogging.info("Detaching node '{}' from network '{}'".format(
                node.name, interface['network']))

            while True:
                try:
                    self.region.ex_destroy_nic(interface['id'])
                    plogging.info("- in progress")

                except Exception as feedback:

                    if 'RESOURCE_BUSY' in str(feedback):
                        time.sleep(10)
                        continue

                    elif 'RESOURCE_LOCKED' in str(feedback):
                        plogging.info("- not now - locked")

                    elif 'NO_CHANGE' in str(feedback):
                        plogging.info("- already there")

                    else:
                        plogging.info("- unable to detach node")
                        plogging.error(str(feedback))
                        return False

                break

        return True

    @classmethod
    def expand_labels(self, label):
        """
        Designates multiple nodes with a simple label

        :param label: the label to be expanded, e.g., ``server[1..2]_eu``
        :type label: ``str``

        :return: a list of names, e.g., ``['server1_eu', 'server2_eu']``
        :rtype: ``list`` of ``str``

        This function creates multiple names where applicable::

            >>>nodes.expand_labels('mongodb')
            ['mongodb']

            >>>nodes.expand_labels('mongodb[1..3]_eu')
            ['mongodb1_eu', 'mongodb2_eu', 'mongodb3_eu']

        """
        matches = re.match(r'(.*)\[([0-9]+)..([0-9]+)\](.*)', label)
        if matches is None:
            if re.match("^[0-9a-zA-Z]([0-9a-zA-Z\-]{0,61}[0-9a-zA-Z])?$", label) is None:
                plogging.warning("Warning: '{}' is not a valid hostname"
                                 .format(label))
            return [label]

        labels = []
        for index in range(int(matches.group(2)), int(matches.group(3))+1):

            label = matches.group(1)+str(index)+matches.group(4)
            if re.match("^[0-9a-zA-Z]([0-9a-zA-Z\-]{0,61}[0-9a-zA-Z])?$", label) is None:
                plogging.warning("Warning: '{}' is not a valid hostname"
                                 .format(label))

            labels.append(label)

        return labels

    @retry(SocketError)
    def get_node(self, path):
        """
        Retrieves a node by name

        :param path: the name of the target node, or its location
        :type path: ``str`` or ``list``of ``str``

        :return: the target node, or None
        :rtype: :class:`libcloud.compute.base.Node`

        This function always make a real API call to get fresh state of the
        target node. Therefore, it can be used in loops where you monitor
        the evolution of the node during build or other change operation.

        This function searches firstly at the current facility. If the
        name is a complete path to a remote node, then plumbery looks
        there. If a different region is provided, then authentication is done
        against the related endpoint.

        For example if ``MyServer`` has been defined in a data centre in
        Europe::

            >>>infrastructure.get_ethernet('MyServer')
            >>>infrastructure.get_ethernet(['EU6', 'MyServer'])
            Looking for remote node 'EU6::MyServer'
            - found it
            >>>infrastructure.get_ethernet(['dd-eu', 'EU6', 'MyServer'])
            Looking for offshore node 'dd-eu::EU6::MyServer'
            - found it


        """

        if isinstance(path, str):
            path = path.split('::')

        node = None

        if len(path) == 2:  # force offshore lookup if needed
            target_region = self.facility.get_region(path[0])
            if target_region != self.facility.get_region():
                path.insert(0, target_region)

        if len(path) == 1:  # local name

            self.facility.power_on()

            for node in self.region.list_nodes():

                if node.extra['datacenterId'] != self.facility.get_location_id():
                    continue

                if node.name == path[0]:

                    self._enrich_node(node)
                    return node

        elif len(path) == 2:  # different location, same region

            self.facility.power_on()

            try:
                self.region.ex_get_location_by_id(path[0])
            except IndexError:
                plogging.warning("'{}' is unknown".format(path[0]))
                return None

            plogging.debug("Looking for remote node '{}'"
                          .format('::'.join(path)))

            for node in self.region.list_nodes():

                if node.extra['datacenterId'] != path[0]:
                    continue

                if node.name == path[1]:

                    plogging.debug("- found it")

                    self._enrich_node(node)
                    return node

        elif len(path) == 3:  # other region

            offshore = self.plumbery.get_compute_driver(region=path[0])

            try:
                remoteLocation = offshore.ex_get_location_by_id(path[1])
            except IndexError:
                plogging.warning("'{}' is unknown".format(path[1]))
                return None

            plogging.debug("Looking for offshore node '{}'"
                          .format('::'.join(path)))

            for node in offshore.list_nodes():

                if node.extra['datacenterId'] != path[1]:
                    continue

                if node.name == path[2]:

                    plogging.debug("- found it")

                    self._enrich_node(node, region=offshore)
                    return node

        return None

    def _enrich_node(self, node, region=None):
        """
        Adds attributes to a node

        This function is a hack, aiming to complement the nice job done by
        Libcloud:
        - add public IPv4 if one exists
        - add disk size, ids, etc.

        """

        if region is None:
            region = self.region

        # hack because the driver does not report public ipv4 accurately
        if len(node.public_ips) < 1:
            domain = region.ex_get_network_domain(
                node.extra['networkDomainId'])
            for rule in self.region.ex_list_nat_rules(domain):
                if rule.internal_ip == node.private_ips[0]:
                    node.public_ips.append(rule.external_ip)
                    break

        # hack to retrieve disk information
        node.extra['disks'] = []
        try:
            element = region.connection.request_with_orgId_api_2(
                'server/server/%s' % node.id).object

            for disk in findall(element, 'disk', TYPES_URN):
                scsiId = int(disk.get('scsiId'))
                speed = disk.get('speed')
                id = disk.get('id')
                sizeGb = int(disk.get('sizeGb'))
                node.extra['disks'].append({
                    'scsiId': scsiId,
                    'speed': speed,
                    'id': id,
                    'size': sizeGb
                    })

        except Exception as feedback:

            if 'RESOURCE_NOT_FOUND' in str(feedback):
                pass

            else:
                plogging.info("Error: unable to retrieve storage information")
                plogging.error(str(feedback))

    @classmethod
    def list_nodes(self, blueprint):
        """
        Retrieves the list of nodes that have been defined for this blueprint.

        :return: names of nodes defined for this blueprint
        :rtype: ``list`` of ``str`` or []

        """

        labels = set()
        if 'nodes' in blueprint:
            for item in blueprint['nodes']:
                if type(item) is dict:
                    label = list(item)[0]

                else:
                    label = str(item)

                for label in PlumberyNodes.expand_labels(label):
                    labels.add(label)

        return list(labels)

    def _list_secondary_interfaces(self, node):
        """
        Retrieves the list of secondary interfaces

        This is a hack. Code here should really go to the Libcloud driver in
        libcloud.compute.drivers.dimensiondata.py _to_node()

        """

        element = self.region.connection.request_with_orgId_api_2(
            'server/server/%s' % node.id).object

        if element.find(fixxpath('networkInfo', TYPES_URN)) is None:
            return []

        interfaces = []
        items = element.findall(
            fixxpath('networkInfo/additionalNic', TYPES_URN))
        for item in items:
            interfaces.append({'id': item.get('id'),
                               'network': item.get('vlanName')})

        return interfaces

    def polish_blueprint(self, blueprint, polishers, container):
        """
        Walks a named blueprint for this facility and polish related resources

        :param blueprint: the blueprint to build
        :type blueprint: ``dict``

        :param polishers: polishers to be applied
        :type polishers: list of :class:`plumbery.PlumberyPolisher`

        :param container: where these nodes are located
        :type container: list of :class:`plumbery.PlumberyInfrastructure`

        """

        if 'nodes' not in blueprint:
            return

        for item in blueprint['nodes']:

            if type(item) is dict:
                label = list(item)[0]
                settings = item[label]
            else:
                label = str(item)
                settings = {}

            for label in self.expand_labels(label):

                node = self.get_node(label)
                settings['name'] = label

                for polisher in polishers:
                    polisher.shine_node(node, settings, container)

    def start_blueprint(self, blueprint):
        """
        Starts nodes of a given blueprint at this facility

        :param blueprint: the blueprint to build
        :type blueprint: ``dict``

        """

        if 'nodes' not in blueprint:
            return

        for item in blueprint['nodes']:

            if type(item) is dict:
                label = list(item)[0]

            else:
                label = item

            for label in self.expand_labels(label):
                self.start_node(label)

    def start_node(self, node):
        """
        Starts one node

        :param node: the target node, or its name
        :type node: :class:`Node` or ``str``

        """

        if isinstance(node, str):
            name = node
            node = self.get_node(name)
        else:
            name = node.name

        plogging.info("Starting node '{}'".format(name))
        if node is None:
            plogging.info("- not found")
            return

        if self.plumbery.safeMode:
            plogging.info("- skipped - safe mode")
            return

        while True:

            try:
                self.region.ex_start_node(node)

                plogging.info("- in progress")

            except Exception as feedback:

                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                elif 'SERVER_STARTED' in str(feedback):
                    plogging.info("- skipped - node is up and running")

                else:
                    plogging.info("- unable to start node")
                    plogging.error(str(feedback))

            break

    def stop_blueprint(self, blueprint):
        """
        Stops nodes of the given blueprint at this facility

        :param blueprint: the blueprint to build
        :type blueprint: ``dict``

        You can use the following setting to prevent plumbery from stopping a
        node::

          - sql:
              domain: *vdc1
              ethernet: *data
              nodes:
                - slaveSQL:
                    running: always

        """

        if 'nodes' not in blueprint:
            return

        for item in blueprint['nodes']:

            if type(item) is dict:
                label = list(item.keys())[0]
                settings = list(item.values())[0]

            else:
                label = item
                settings = {}

            for label in self.expand_labels(label):
                self.stop_node(label, settings)

    def stop_node(self, node, settings={}):
        """
        Stops one node

        :param node: the target node, or its name
        :type node: :class:`Node` or ``str``

        :param settings: additional attributes for this node
        :type settings: ``dict``

        """

        if isinstance(node, str):
            name = node
            node = self.get_node(name)
        else:
            name = node.name

        plogging.info("Stopping node '{}'".format(name))
        if node is None:
            plogging.info("- not found")
            return

        if ('running' in settings
                and settings['running'] == 'always'
                and node.state == NodeState.RUNNING):

            plogging.info("- skipped - node has to stay always on")
            return

        if self.plumbery.safeMode:
            plogging.info("- skipped - safe mode")
            return

        retry = True
        while True:

            try:
                self.region.ex_shutdown_graceful(node)
                plogging.info("- in progress")

            except Exception as feedback:

                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                elif 'UNEXPECTED_ERROR' in str(feedback):
                    time.sleep(10)
                    continue

                elif 'VMWARE_TOOLS_INVALID_STATUS' in str(feedback):

                    # prevent transient errors
                    if retry:
                        retry = False
                        time.sleep(30)
                        continue

                    plogging.info("- unable to shutdown gracefully "
                                  "- invalid VMware tools")

                    plogging.info("- powering the node off")
                    try:
                        self.region.ex_power_off(node)
                        plogging.info("- in progress")

                    except Exception as feedback:

                        if 'SERVER_STOPPED' in str(feedback):
                            plogging.info("- already stopped")

                        else:
                            plogging.info("- unable to stop node")
                            plogging.error(str(feedback))

                elif 'SERVER_STOPPED' in str(feedback):
                    plogging.info("- already stopped")

                else:
                    plogging.info("- unable to stop node")
                    plogging.error(str(feedback))

            break



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

from plumbery.polishers.base import NodeConfiguration
from plumbery.exception import ConfigurationError
from plumbery.plogging import plogging


class DisksConfiguration(NodeConfiguration):
    __name__ = 'DisksConfiguration'
    _element_name_ = 'disks'
    _configuration_ = {
    }

    def __init__(self, engine, facility):
        self.engine = engine
        self.facility = facility

    def validate(self, settings):
        if self._element_name_ in settings:
            for item in settings[self._element_name_]:
                tokens = item.lower().split()
                if len(tokens) < 2 or len(tokens) > 3:
                    raise ConfigurationError(
                        "- malformed disk attributes;"
                        " provide disk id and size in GB, e.g., 1 50;"
                        " add disk type if needed, e.g., economy")
                if len(tokens) < 3:
                    tokens.append('standard')

                if int(tokens[0]) not in (0, 1, 2, 3, 4, 5, 6, 7, 8, 9):
                    raise ConfigurationError(
                        "- disk id should be between 0 and 9")

                if int(tokens[1]) < 10:
                    raise ConfigurationError(
                        "- minimum disk size is 10 GB")

                if int(tokens[1]) > 1000:
                    raise ConfigurationError(
                        "- disk size cannot exceed 1000 GB")

                if tokens[2] not in ('standard', 'highperformance', 'economy'):
                    raise ConfigurationError(
                        "- disk speed should be either 'standard' "
                         "or 'highperformance' or 'economy'")
        return True

    def configure(self, node, settings):
        if self._element_name_ in settings:
            for item in settings['disks']:
                plogging.debug("- setting disk {}".format(item))
                attributes = item.split()
                if len(attributes) < 2:
                    plogging.info("- malformed disk attributes;"
                                 " provide disk id and size in GB, e.g., 1 50;"
                                 " add disk type if needed, e.g., economy")
                elif len(attributes) < 3:
                    id = int(attributes[0])
                    size = int(attributes[1])
                    speed = 'standard'
                else:
                    id = int(attributes[0])
                    size = int(attributes[1])
                    speed = attributes[2]

                self.set_node_disk(node, id, size, speed)
            return True
        return False

    def deconfigure(self, node, settings):
        return True

    def set_node_disk(self, node, id, size, speed='standard'):
        """
        Sets a virtual disk

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param id: the disk id, starting at 0 and growing
        :type id: ``int``

        :param size: the disk size, expressed in Giga bytes
        :type size: ``int``

        :param speed: storage type, either 'standard',
            'highperformance' or 'economy'
        :type speed: ``str``

        """

        if size < 1:
            plogging.info("- minimum disk size is 1 GB")
            return

        if size > 1000:
            plogging.info("- disk size cannot exceed 1000 GB")
            return

        if speed not in ('standard', 'highperformance', 'economy'):
            plogging.info("- disk speed should be either 'standard' "
                         "or 'highperformance' or 'economy'")
            return

        if 'disks' in node.extra:
            for disk in node.extra['disks']:
                if disk['scsiId'] == id:
                    changed = False

                    if disk['size'] > size:
                        plogging.info("- disk shrinking could break the node")
                        plogging.info("- skipped - disk {} will not be reduced"
                                     .format(id))

                    if disk['size'] < size:
                        plogging.info("- expanding disk {} to {} GB"
                                     .format(id, size))
                        self.change_node_disk_size(node, disk['id'], size)
                        changed = True

                    if disk['speed'].lower() != speed.lower():
                        plogging.info("- changing disk {} to '{}'"
                                     .format(id, speed))
                        self.change_node_disk_speed(node, disk['id'], speed)
                        changed = True

                    if not changed:
                        plogging.debug("- no change in disk {}".format(id))

                    return

        plogging.info("- adding {} GB '{}' disk".format(
            size, speed))

#        if self.engine.safeMode:
#            plogging.info("- skipped - safe mode")
#            return

        while True:
            try:
                self.facility.region.ex_add_storage_to_node(
                    node=node,
                    amount=size,
                    speed=speed.upper())

                plogging.info("- in progress")

            except Exception as feedback:
                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                if 'Please try again later' in str(feedback):
                    time.sleep(10)
                    continue

                plogging.info("- unable to add disk {} GB '{}'"
                             .format(size, speed))
                plogging.error(str(feedback))

            break

    def change_node_disk_size(self, node, id, size):
        """
        Changes an existing virtual disk

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param id: the disk unique identifier, as reported by the API
        :type id: ``str``

        :param size: the disk size, expressed in Giga bytes
        :type size: ``int``

        """

        if self.engine.safeMode:
            plogging.info("- skipped - safe mode")
            return

        while True:
            try:
                self.facility.region.ex_change_storage_size(
                    node=node,
                    disk_id=id,
                    size=size)

                plogging.info("- in progress")

            except Exception as feedback:
                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                if 'Please try again later' in str(feedback):
                    time.sleep(10)
                    continue

                plogging.info("- unable to change disk size to {}GB"
                             .format(size))
                plogging.error(str(feedback))

            break

    def change_node_disk_speed(self, node, id, speed):
        """
        Changes an existing virtual disk

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param id: the disk unique identifier, as reported by the API
        :type id: ``str``

        :param speed: storage type, either 'standard',
            'highperformance' or 'economy'
        :type speed: ``str``

        """

        if self.engine.safeMode:
            plogging.info("- skipped - safe mode")
            return

        while True:
            try:
                self.facility.region.ex_change_storage_speed(
                    node=node,
                    disk_id=id,
                    speed=speed)

                plogging.info("- in progress")

            except Exception as feedback:
                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                if 'Please try again later' in str(feedback):
                    time.sleep(10)
                    continue

                plogging.info("- unable to change disk to '{}'"
                             .format(speed))
                plogging.error(str(feedback))

            break


# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# this is required because we called this file logging.py
#
from __future__ import absolute_import

import logging
import colorlog


class PlumberyLogging(object):

    def __init__(self):

        self.logger = logging.getLogger('plumbery')
        self.logger.propagate = 0

        # logging to console
        #
        handler = colorlog.StreamHandler()
        formatter = colorlog.ColoredFormatter(
            "%(asctime)-2s %(log_color)s%(message)s",
            datefmt='%H:%M:%S',
            reset=True,
            log_colors={
                'DEBUG':    'cyan',
                'INFO':     'green',
                'WARNING':  'yellow',
                'ERROR':    'red',
                'CRITICAL': 'red,bg_white',
            },
            secondary_log_colors={},
            style='%'
        )
        handler.setFormatter(formatter)

        self.logger.addHandler(handler)

        logging.getLogger('paramiko.transport').addHandler(handler)
        logging.getLogger('paramiko.transport').propagate = 0

        self.reset()

    def reset(self):
        self.errors = 0

    def foundErrors(self):
        return self.errors > 0

    def debug(self, *args):
        self.logger.debug(*args)

    def info(self, *args):
        self.logger.info(*args)

    def warning(self, *args):
        self.logger.warning(*args)

    def error(self, *args):
        self.logger.error(*args)
        self.errors += 1

    def critical(self, *args):
        self.logger.critical(*args)
        self.errors += 1

    def getEffectiveLevel(self):
        return self.logger.getEffectiveLevel()

    def setLevel(self, level):
        self.logger.setLevel(level)

    def addHandler(self, handler):
        self.logger.addHandler(handler)

plogging = PlumberyLogging()

logging.basicConfig(level=logging.INFO)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import requests

from pywinexe.api import cmd as run_cmd
from pywinexe.api import ps as run_ps

from libcloud.compute.types import NodeState
import winrm
from winrm.protocol import Protocol

from plumbery.polishers.base import NodeConfiguration
from plumbery.plogging import plogging


class WindowsConfiguration(NodeConfiguration):
    __name__ = 'WindowsConfiguration'
    _element_name_ = 'windows'
    _configuration_ = {
    }

    def __init__(self, engine, facility):
        self.secret = engine.get_shared_secret()
        # todo: provide a fittings-wide override.
        self.username = 'administrator'
        plogging.debug('Loading windows polisher')

    def _try_winrm(self, node):
        ip = node.private_ips[0]
        p = Protocol(
                endpoint='http://%s:5985/wsman' % ip,  # RFC 2732
                transport='ntlm',
                username=self.username,
                password=self.secret,
                server_cert_validation='ignore')
        shell_id = p.open_shell()
        command_id = p.run_command(shell_id, 'ipconfig', ['/all'])
        std_out, std_err, status_code = p.get_command_output(shell_id, command_id)
        p.cleanup_command(shell_id, command_id)
        p.close_shell(shell_id)
        return std_out

    def _winrm_commands(self, node, commands):
        ip = node.private_ips[0]
        p = Protocol(
                endpoint='http://%s:5985/wsman' % ip,  # RFC 2732
                transport='ntlm',
                username=self.username,
                password=self.secret,
                server_cert_validation='ignore')
        shell_id = p.open_shell()
        std_out_logs = []
        std_err_logs = []

        # run the commands in sequence.
        for command, params in commands:
            command_id = p.run_command(shell_id, command, params)
            std_out, std_err, status_code = p.get_command_output(shell_id, command_id)
            std_out_logs.append(std_out)
            std_err_logs.append(std_err)
            p.cleanup_command(shell_id, command_id)

        p.close_shell(shell_id)
        return std_out_logs, std_err_logs

    def _setup_winrm(self, node):
        """
        Setup WinRM on a remote node

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`
        """
        ip = node.private_ips[0]
        plogging.debug("Testing out quick function on %s", ip)
        out = run_cmd(
            'echo hello',
            args=[],
            user=self.username,
            password=self.secret,
            host=ip)
        plogging.info(out)
        plogging.debug("Running winexe to remotely configure %s", ip)
        cmds = [
            "winrm quickconfig -quiet",
            "winrm set winrm/config/service/auth @{Basic=\"true\"}",
            "winrm set winrm/config/service @{AllowUnencrypted=\"true\"}"
        ]
        for cmd in cmds:
            plogging.debug('Running command "%s"', cmd)
            out = run_cmd(
                cmd,
                args=[],
                user=self.username,
                password=self.secret,
                host=ip)
            plogging.info(out)

    def _lockdown_winrm(self, node):
        """
        Setup WinRM on a remote node

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`
        """
        ip = node.private_ips[0]
        plogging.debug("Running winexe to remotely deconfigure %s", ip)
        cmds = [
            "winrm set winrm/config/service/auth @{Basic=\"false\"}",
            "winrm set winrm/config/service @{AllowUnencrypted=\"false\"}"
        ]
        for cmd in cmds:
            plogging.debug('Running command "%s"', cmd)
            out = run_cmd(
                cmd,
                args=[],
                user=self.username,
                password=self.secret,
                host=ip)
            plogging.info(out)

    def validate(self, settings):
        return True

    def reap(self, *args):
        plogging.debug('Reap for windows polisher (noop)')
        return

    def configure(self, node, settings):
        """
        prepares a node

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param settings: the fittings plan for this node
        :type settings: ``dict``

        :param container: the container of this node
        :type container: :class:`plumbery.PlumberyInfrastructure`

        """
        if self._element_name_ in settings:
            plogging.info("preparing node '{}'".format(settings['name']))
            if node is None:
                plogging.info("- not found")
                return

            timeout = 300
            tick = 6
            while node.extra['status'].action == 'START_SERVER':
                time.sleep(tick)
                node = self.nodes.get_node(node.name)
                timeout -= tick
                if timeout < 0:
                    break

                if node.state != NodeState.RUNNING:
                    plogging.info("- skipped - node is not running")
                    return

            ipv6 = node.extra['ipv6']
            ip = node.private_ips[0]
            if ipv6 is None:
                plogging.error('No ipv6 address for node, cannot configure')
                return

            # Check to see if WinRM works..
            try:
                self._try_winrm(node)
            except winrm.exceptions.InvalidCredentialsError:
                plogging.warn('initial login to %s failed, trying to setup winrm remotely',
                             ip)
                self._setup_winrm(node)
                self._try_winrm(node)
            except requests.exceptions.ConnectionError:
                plogging.warn('initial connection to %s failed, trying to setup winrm remotely',
                             ip)
                self._setup_winrm(node)
                self._try_winrm(node)

            # OK, we're all ready. Let's look at the node config and start commands
            cmds = []
            hostname = settings[self._element_name_].get('hostname', None)
            if hostname is not None and isinstance(hostname, str):
                cmds.append(('powershell.exe', ['Rename-Computer', '-NewName', hostname]))

            extra_cmds = settings[self._element_name_].get('cmds', [])
            for command in extra_cmds:
                command = command.rstrip()
                command_parts = command.split(' ')
                cmds.append((command_parts[0], command_parts[1:]))

            out, err = self._winrm_commands(node, cmds)
            plogging.info(out)
            plogging.warning(err)

            plogging.debug('locking down winrm')
            self._lockdown_winrm(node)
        else:
            return False



#!/usr/bin/env python

"""
Tests for `infrastructure` module.
"""

# special construct to allow relative import
#
if __name__ == "__main__" and __package__ is None:
    __package__ = "tests"
from tests import dummy

import os
import unittest

from libcloud.compute.drivers.dimensiondata import DimensionDataNodeDriver
from libcloud.backup.drivers.dimensiondata import DimensionDataBackupDriver

from plumbery.engine import PlumberyEngine
from plumbery.infrastructure import PlumberyInfrastructure

from .mock_api import DimensionDataMockHttp
DIMENSIONDATA_PARAMS = ('user', 'password')


class FakeDomain:

    id = 123


class FakeNetwork:

    id = 123


class FakePlumbery:

    safeMode = False
    working_directory = os.getcwd()
    def get_balancer_driver(self, region):
        return None


class FakeLocation:

    id = 'EU6'


fakeParameters = {
    'regionId': 'dd-na',
    'locationId': 'NA9'
}

fakeBlueprints = [{
    'fake': {
        'domain': {
            'name': 'VDC1',
            'service': 'ADVANCED',
            'description': 'fake'},
        'ethernet': {
            'name': 'vlan1',
            'subnet': '10.0.10.0',
            'description': 'fake'},
        'nodes': [{
            'stackstorm': {
                'description': 'fake',
                'appliance': 'RedHat 6 64-bit 4 CPU'
                }
            }]
        }
    }]


class FakeFacility:

    plumbery = FakePlumbery()

    settings = fakeParameters
    blueprints = fakeBlueprints

    DimensionDataNodeDriver.connectionCls.conn_classes = (
        None, DimensionDataMockHttp)
    DimensionDataBackupDriver.connectionCls.conn_classes = (
        None, DimensionDataMockHttp)
    DimensionDataMockHttp.type = None
    region = DimensionDataNodeDriver(*DIMENSIONDATA_PARAMS)
    backup = DimensionDataBackupDriver(*DIMENSIONDATA_PARAMS)
    location = FakeLocation()

    _cache_network_domains = []
    _cache_vlans = []

    def get_location_id(self):
        return 'EU6'

    def get_setting(self, label):
        if label in self.settings:
            return self.settings[label]

        return None

fakeBluePrint = {'target': 'fake',
                 'domain': {'name': 'fake',
                            'service': 'ADVANCED',
                            'description': '#vdc1'},
                 'ethernet': {'name': 'fake',
                              'subnet': '10.0.10.0',
                              'description': '#vdc1'}}

defaultsPlan = """
---
safeMode: True
defaults:
  locationId: EU6
  regionId: dd-eu
  ipv4: auto
  cloud-config:
    disable_root: false
    ssh_pwauth: true
    ssh_keys:
      rsa_private: |
        {{ pair1.rsa_private }}

      rsa_public: "{{ pair1.ssh.rsa_public }}"

---

basement: central

blueprints:

  - myBlueprint:
      domain:
        name: myDC
      ethernet:
        name: myVLAN
        subnet: 10.1.10.0
      nodes:
        - myServer:
            glue:
              - primary .11
              - centralVLAN .12

  - central:
      domain:
        name: myDC
      ethernet:
        name: centralVLAN
        subnet: 10.2.10.0
"""


class TestPlumberyInfrastructure(unittest.TestCase):

    def setUp(self):
        facility = FakeFacility()
        self.infrastructure = PlumberyInfrastructure(facility=facility)

    def tearDown(self):
        self.infrastructure = None

    def test_parse_firewall_port(self):

        elements = self.infrastructure.parse_firewall_port('icmp')
        self.assertEqual(elements, ('ICMP', 'any', None, None))

        elements = self.infrastructure.parse_firewall_port('tcp:80')
        self.assertEqual(elements, ('TCP', '80', '80', None))

        elements = self.infrastructure.parse_firewall_port(':80')
        self.assertEqual(elements, ('TCP', '80', '80', None))

        elements = self.infrastructure.parse_firewall_port('80')
        self.assertEqual(elements, ('TCP', '80', '80', None))

        elements = self.infrastructure.parse_firewall_port('udp:137..138')
        self.assertEqual(elements, ('UDP', '137..138', '137', '138'))

        elements = self.infrastructure.parse_firewall_port('any')
        self.assertEqual(elements, ('TCP', 'any', None, None))

    def test_build(self):
        self.infrastructure.build(fakeBluePrint)

    def test_name_balancer(self):
        self.infrastructure.blueprint = fakeBluePrint
        name = self.infrastructure.name_balancer('fake')
        self.assertEqual(name, 'fake.fake.eu6.balancer')

    def test_get_balancer(self):
        self.infrastructure.blueprint = fakeBluePrint
        balancer = self.infrastructure._get_balancer('fakeListener')
        self.assertEqual(balancer, None)

    def test_name_pool(self):
        self.infrastructure.blueprint = fakeBluePrint
        name = self.infrastructure._name_pool()
        self.assertEqual(name, 'fake.eu6.pool')

    def test_get_pool(self):
        self.infrastructure.blueprint = fakeBluePrint
        pool = self.infrastructure._get_pool()
        self.assertEqual(pool, None)

    def test_get_container(self):
        container = self.infrastructure.get_container(fakeBluePrint)
        self.assertEqual(container.domain, None)
        self.assertEqual(container.network, None)

    def test_get_ethernet(self):
        self.infrastructure.get_ethernet('MyNetwork')
#        self.infrastructure.get_ethernet(['XY6', 'MyNetwork'])
#        self.infrastructure.get_ethernet(['dd-eu', 'EU6', 'MyNetwork'])

    def test_get_ipv4(self):
        self.infrastructure.blueprint = fakeBluePrint
        self.infrastructure._get_ipv4()

    def test_get_default(self):
        engine = PlumberyEngine(defaultsPlan)
        facility = engine.list_facility('EU6')[0]
        infrastructure = PlumberyInfrastructure(facility)
        self.assertEqual(infrastructure.get_default('ipv4'), 'auto')

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python

"""
Tests for `polisher` module.
"""

# special construct to allow relative import
#
if __name__ == "__main__" and __package__ is None:
    __package__ = "tests"
from tests import dummy

from collections import namedtuple
import unittest

from libcloud.compute.drivers.dimensiondata import DimensionDataNodeDriver
from libcloud.compute.types import NodeState

from plumbery.engine import PlumberyEngine
from plumbery.infrastructure import PlumberyInfrastructure
from plumbery.nodes import PlumberyNodes
from plumbery.polisher import PlumberyPolisher

from .mock_api import DimensionDataMockHttp
DIMENSIONDATA_PARAMS = ('user', 'password')


class FakeNetwork:

    id = 123


class FakeEngine():

    information = []

    def get_shared_secret(self):
        return 'nuts'

    def get_default(self, label, default=None):
        return default


class FakeRegion:

    def create_node(self, name, image, auth, ex_network_domain, ex_vlan,
                    ex_is_started, ex_description):
        return True

    def ex_create_vlan(self, network_domain, name, private_ipv4_base_address,
                       description):
        return FakeNetwork()

    def ex_get_network_domain(self, location, network_domain):
        return []

    def ex_get_vlan(self, vlan_id):
        return FakeNetwork()

    def ex_list_nat_rules(self, domain):
        return []

    def ex_list_network_domains(self, location):
        return []

    def ex_list_vlans(self, location):
        return []

    def ex_start_node(self, node):
        return True

    def ex_shutdown_graceful(self, node):
        return True

    def ex_wait_for_state(self, state, func, poll_interval=2, timeout=60,
                          *args, **kwargs):
        return []

    def destroy_node(self, node):
        return True

    def list_nodes(self):
        return [FakeNode()]


fakeFacilitySettings = {
    'locationId': 'EU6',
    'prepare': [{'beachhead': '10.1.10.9'}, {'beachhead': '10.1.10.10'}],
    'regionId': 'dd-eu'}


class FakeFacility():
    plumbery = FakeEngine()
    settings = fakeFacilitySettings
    region = FakeRegion()
    backup = None

    def __repr__(self):
        return "<FakeFacility fittings: {}>".format(self.fittings)

    def get_location_id(self):
        return 'EU6'

    def get_setting(self, label, default=None):
        if label in self.settings:
            return self.settings[label]

        return default

    def power_on(self):
        pass


class FakeContainer:

    id = 123
    domain = 'fake'
    network = FakeNetwork()

    region = FakeRegion()
    facility = FakeFacility()

    blueprint = {
        'target': 'fake',
        'domain': {
            'name': 'VDC1',
            'service': 'ADVANCED',
            'description': 'fake'},
        'ethernet': {
            'name': 'vlan1',
            'subnet': '10.0.10.0',
            'description': 'fake'},
        'nodes': [{
            'stackstorm': {
                'description': 'fake',
                'appliance': 'RedHat 6 64-bit 4 CPU'
                }
            }]
        }

    def get_network_domain(self, blueprint):
        return None

    def get_ethernet(self, blueprint):
        return None

    def _add_to_pool(self, node):
        pass

FakeStatus = namedtuple('FakeStatus', 'action')


class FakeNode():

    def __init__(self):
        self.name = 'fake'
        self.id = '1234'
        self.state = NodeState.RUNNING
        self.private_ips = ['12.34.56.78']
        self.public_ips = []
        self.extra = {'datacenterId': 'EU6',
                      'description': '#fake description with #tags',
                      'ipv6': 'fe80::', 'here': 'there',
                      'status': FakeStatus('none'),
                      'dummy': 'test'}

fakeNodeSettings = {
    'name': 'stackstorm',
    'description': 'fake',
    'appliance': 'RedHat 6 64-bit 4 CPU',
    'information': ['hello world'],
    'prepare': ['prepare.update.sh', 'prepare.docker.sh']}

fakePrepareConfiguration = {
    'key': 'test_polisher.pub'}


myInformation = """
---
information:
  - plan-level information
---
locationId: NA9
regionId: dd-na
information:
  - facility-level information

blueprints:

  - test:
      domain:
        description: fake
        name: myDC
      ethernet:
        description: fake
        name: myVLAN
        subnet: 10.1.10.0
      information:
        - container-level information
      nodes:
        - stackstorm:
            description: fake
            appliance: 'RedHat 6 64-bit 4 CPU'
            cpu: 2
            memory: 2
            disks:
              - 1 10 highperformance
            glue:
              - internet 22
            monitoring: essentials
            information:
              - hello world
            prepare:
              - prepare.update.sh
              - prepare.docker.sh
        - node1:
            information:
              - node-level information
"""


def do_polish(polisher):

    engine = PlumberyEngine(myInformation)
    engine.set_shared_secret('fake_secret')
    engine.set_user_name('fake_name')
    engine.set_user_password('fake_password')

    polisher.go(engine)

    facility = engine.list_facility('NA9')[0]
    DimensionDataNodeDriver.connectionCls.conn_classes = (
        None, DimensionDataMockHttp)
    DimensionDataMockHttp.type = None
    facility.region = DimensionDataNodeDriver(*DIMENSIONDATA_PARAMS)

    polisher.move_to(facility)

    blueprint = facility.get_blueprint('test')
    infrastructure = PlumberyInfrastructure(facility)
    container = infrastructure.get_container(blueprint)

    polisher.shine_container(container)

    nodes = PlumberyNodes(facility)

    node = nodes.get_node('stackstorm')
    polisher.shine_node(
        node=node, settings=fakeNodeSettings, container=container)

    node = nodes.get_node('node1')
    polisher.shine_node(
        node=node, settings=fakeNodeSettings, container=container)

    polisher.move_to(FakeFacility())

    polisher.shine_container(FakeContainer())

    polisher.shine_node(
        node=FakeNode(), settings=fakeNodeSettings, container=FakeContainer())

    polisher.reap()


class TestPlumberyPolisher(unittest.TestCase):

    def test_ansible(self):
        polisher = PlumberyPolisher.from_shelf('ansible', {})
        do_polish(polisher)

    def test_configure(self):
        polisher = PlumberyPolisher.from_shelf('configure', {})
        do_polish(polisher)

    def test_information(self):
        polisher = PlumberyPolisher.from_shelf('information', {})
        do_polish(polisher)

    def test_inventory(self):
        polisher = PlumberyPolisher.from_shelf('inventory', {})
        do_polish(polisher)

    def test_ping(self):
        polisher = PlumberyPolisher.from_shelf('ping', {})
        do_polish(polisher)

    def test_prepare(self):
        polisher = PlumberyPolisher.from_shelf(
            'prepare', fakePrepareConfiguration)
        do_polish(polisher)

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python

"""
Tests for `fittings` module.
"""

import unittest

from plumbery.fitting import PlumberyFitting, PlumberyFittingLoader


class TestPlumberyFitting(unittest.TestCase):

    def test_loader_unknown_class(self):

        with self.assertRaises(ImportError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='*123*',
                engine=None,
                facility=None,
                settings={'dummy': 'no'})

    def test_dummy(self):

        fitting = PlumberyFittingLoader.from_shelf(
            label='dummy',
            engine=None,
            facility=None,
            settings={'dummy': 'no'})
        self.assertTrue(isinstance(fitting, PlumberyFitting))
        self.assertFalse(fitting.completed)
        self.assertTrue(fitting.do('unmanaged_action'))
        self.assertFalse(fitting.completed)
        self.assertTrue(fitting.do('some_action'))
        self.assertTrue(fitting.completed)

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='dummy',
                engine=None,
                facility=None,
                settings=('dummy', 'no'))

        with self.assertRaises(KeyError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='dummy',
                engine=None,
                facility=None,
                settings={})

        with self.assertRaises(KeyError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='dummy',
                engine=None,
                facility=None,
                settings={'*123*': 'no'})

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='dummy',
                engine=None,
                facility=None,
                settings={'dummy': True})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='dummy',
                engine=None,
                facility=None,
                settings={'dummy': ''})

    def test_domain(self):

        # minimum viable settings

        fitting = PlumberyFittingLoader.from_shelf(
            label='domain',
            engine=None,
            facility=None,
            settings={'name': 'vdc1'})
        self.assertTrue(isinstance(fitting, PlumberyFitting))
        self.assertEqual(fitting.label, 'domain')
        self.assertEqual(fitting.description, '#plumbery')
        self.assertEqual(fitting.ipv4, None)
        self.assertEqual(fitting.name, 'vdc1')
        self.assertEqual(fitting.service, 'ESSENTIALS')

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings=('domain', 'ipv4'))

        with self.assertRaises(KeyError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'*123*': 'no'})

        # description

        fitting = PlumberyFittingLoader.from_shelf(
            label='domain',
            engine=None,
            facility=None,
            settings={'name': 'vdc1', 'description': 'hello'})
        self.assertEqual(fitting.description, 'hello #plumbery')

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'name': 'vdc1', 'description': True})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'name': 'vdc1', 'description': ''})

        # ipv4

        fitting = PlumberyFittingLoader.from_shelf(
            label='domain',
            engine=None,
            facility=None,
            settings={'name': 'vdc1', 'ipv4': 'auto'})
        self.assertEqual(fitting.ipv4, 'auto')

        fitting = PlumberyFittingLoader.from_shelf(
            label='domain',
            engine=None,
            facility=None,
            settings={'name': 'vdc1', 'ipv4': 6})
        self.assertEqual(fitting.ipv4, 6)

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'name': 'vdc1', 'ipv4': True})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'name': 'vdc1', 'ipv4': -2})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'name': 'vdc1', 'ipv4': 321})

        # name

        with self.assertRaises(KeyError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'description': 'test', 'ipv4': 6})

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'name': True, 'ipv4': 'auto'})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'name': ''})

        # service

        fitting = PlumberyFittingLoader.from_shelf(
            label='domain',
            engine=None,
            facility=None,
            settings={'name': 'vdc1', 'service': 'essentials'})
        self.assertEqual(fitting.service, 'ESSENTIALS')

        fitting = PlumberyFittingLoader.from_shelf(
            label='domain',
            engine=None,
            facility=None,
            settings={'name': 'vdc1', 'service': 'ESSENTIALS'})
        self.assertEqual(fitting.service, 'ESSENTIALS')

        fitting = PlumberyFittingLoader.from_shelf(
            label='domain',
            engine=None,
            facility=None,
            settings={'name': 'vdc1', 'service': 'advanced'})
        self.assertEqual(fitting.service, 'ADVANCED')

        fitting = PlumberyFittingLoader.from_shelf(
            label='domain',
            engine=None,
            facility=None,
            settings={'name': 'vdc1', 'service': 'ADVANCED'})
        self.assertEqual(fitting.service, 'ADVANCED')

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'name': 'vdc1', 'service': True})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='domain',
                engine=None,
                facility=None,
                settings={'name': 'vdc1', 'service': '*123*'})

    def test_ethernet(self):

        # minimum viable settings

        fitting = PlumberyFittingLoader.from_shelf(
            label='ethernet',
            engine=None,
            facility=None,
            settings={'name': 'vlan1', 'subnet': '10.2.3.0'})
        self.assertTrue(isinstance(fitting, PlumberyFitting))
        self.assertEqual(fitting.label, 'ethernet')
        self.assertEqual(fitting.description, '#plumbery')
        self.assertEqual(fitting.name, 'vlan1')
        self.assertEqual(fitting.subnet, '10.2.3.0')

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings=('description', 'subnet'))

        with self.assertRaises(KeyError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings={'*123*': 'no'})

        # description

        fitting = PlumberyFittingLoader.from_shelf(
            label='ethernet',
            engine=None,
            facility=None,
            settings={'name': 'vlan1', 'description': 'hello', 'subnet': '10.2.3.0'})
        self.assertEqual(fitting.description, 'hello #plumbery')

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings={'name': 'vlan1', 'description': True, 'subnet': '10.2.3.0'})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings={'name': 'vlan1', 'description': '', 'subnet': '10.2.3.0'})

        # name

        with self.assertRaises(KeyError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings={'description': 'test', 'subnet': '10.2.3.0'})

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings={'name': True, 'subnet': '10.2.3.0'})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings={'name': '', 'subnet': '10.2.3.0'})

        # subnet

        fitting = PlumberyFittingLoader.from_shelf(
            label='ethernet',
            engine=None,
            facility=None,
            settings={'name': 'vlan1', 'subnet': '10.2.3.0'})
        self.assertEqual(fitting.subnet, '10.2.3.0')

        fitting = PlumberyFittingLoader.from_shelf(
            label='ethernet',
            engine=None,
            facility=None,
            settings={'name': 'vlan1', 'subnet': '192.6.7.0'})
        self.assertEqual(fitting.subnet, '192.6.7.0')

        with self.assertRaises(TypeError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings={'name': 'vlan1', 'subnet': True})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings={'name': 'vlan1', 'subnet': '*123*'})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings={'name': 'vlan1', 'subnet': '8.8.8'})

        with self.assertRaises(ValueError):
            fitting = PlumberyFittingLoader.from_shelf(
                label='ethernet',
                engine=None,
                facility=None,
                settings={'name': 'vlan1', 'subnet': '8.8.8.8.8'})

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.compute.base import NodeState

from plumbery.polisher import PlumberyPolisher
from plumbery.plogging import plogging


class PingPolisher(PlumberyPolisher):
    """
    Checks state of nodes

    """

    def shine_node(self, node, settings, container):
        """
        Lists network address of a node

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`

        :param settings: the fittings plan for this node
        :type settings: ``dict``

        :param container: the container of this node
        :type container: :class:`plumbery.PlumberyInfrastructure`

        """

        plogging.info("Pinging node '{}'".format(settings['name']))
        if node is None:
            plogging.info("- not found")
            return

        if 'description' in node.extra:
            description = node.extra['description'].replace(
                '#plumbery', '').strip()
            if len(description) > 0:
                plogging.info("- {}".format(description))

        if node.state == NodeState.RUNNING:
            plogging.info("- node is up and running")
        elif node.state in [NodeState.TERMINATED,
                            NodeState.STOPPED,
                            NodeState.SUSPENDED]:
            plogging.info("- node has been stopped")
        else:
            plogging.info("- state: {}".format(node.state))

        # hack because the driver does not report public ipv4 accurately
        if len(node.public_ips) < 1:
            domain = container.get_network_domain(
                container.blueprint['domain']['name'])
            for rule in container.region.ex_list_nat_rules(domain):
                if rule.internal_ip == node.private_ips[0]:
                    node.public_ips.append(rule.external_ip)
                    break

        if len(node.public_ips) > 0:
            plogging.info("- public: {}".format(
                node.public_ips[0]))

        for item in node.private_ips:
            plogging.info("- private: {}".format(str(item)))

        if node.extra['ipv6']:
            plogging.info("- ipv6: {}".format(str(node.extra['ipv6'])))



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from libcloud.compute.types import NodeState

from plumbery.polishers.base import NodeConfiguration
from plumbery.exception import ConfigurationError
from plumbery.plogging import plogging


class MonitoringConfiguration(NodeConfiguration):
    __name__ = 'MonitoringConfiguration'
    _element_name_ = 'monitoring'
    _configuration_ = {
    }

    def __init__(self, engine, facility):
        self.engine = engine
        self.facility = facility

    def validate(self, settings):
        if self._element_name_ in settings:
            value = settings[self._element_name_].upper()
            if value not in ['ESSENTIALS', 'ADVANCED']:
                raise ConfigurationError(
                    "- monitoring should be "
                    "either 'essentials' or 'advanced'")
        return True

    def configure(self, node, settings):
        if self._element_name_ in settings:
            self._start_monitoring(node, settings[self._element_name_].upper())
            return True
        return False

    def deconfigure(self, node, settings):
        if self._element_name_ in settings:
            self._stop_monitoring(node, settings)
            return True
        return False

    def _start_monitoring(self, node, monitoring='ESSENTIALS'):
        """
        Enables monitoring of one node

        :param node: the target node
        :type node: :class:`libcloud.compute.base.Node`

        :param monitoring: either 'ESSENTIALS' or 'ADVANCED'
        :type monitoring: ``str``

        """

        value = monitoring.upper()
        plogging.info("Starting {} monitoring of node '{}'".format(
            value.lower(), node.name))

        while True:
            try:
                self.facility.region.ex_enable_monitoring(node, service_plan=value)
                plogging.info("- in progress")
                return True

            except Exception as feedback:
                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                elif 'RETRYABLE_SYSTEM_ERROR' in str(feedback):
                    time.sleep(10)
                    continue

                elif 'NO_CHANGE' in str(feedback):
                    plogging.info("- already there")

                elif 'RESOURCE_LOCKED' in str(feedback):
                    plogging.info("- unable to start monitoring "
                                 "- node has been locked")

                else:
                    plogging.info("- unable to start monitoring")
                    plogging.error(str(feedback))

            break

        return False

    def _stop_monitoring(self, node, settings):
        """
        Disables monitoring of one node

        :param node: the target node
        :type node: :class:`libcloud.compute.base.Node`

        """

        if node is None:
            return

        if ('running' in settings
                and settings['running'] == 'always'
                and node.state == NodeState.RUNNING):

            return

        if self.engine.safeMode:
            return

        plogging.info("Stopping monitoring of node '{}'".format(node.name))

        while True:

            try:
                self.facility.region.ex_disable_monitoring(node)
                plogging.info("- in progress")

            except Exception as feedback:

                if 'NO_CHANGE' in str(feedback):
                    pass

                elif 'OPERATION_NOT_SUPPORTED' in str(feedback):
                    pass

                elif 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                elif 'RESOURCE_NOT_FOUND' in str(feedback):
                    plogging.info("- not found")

                elif 'RESOURCE_LOCKED' in str(feedback):
                    plogging.info("- not now - locked")

                else:
                    plogging.info("- unable to stop monitoring")
                    plogging.error(str(feedback))

            break



#!/usr/bin/env python

"""
Tests for `util` module.

Credit of test for the @retry decorator:
http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
"""

import logging
import unittest

from plumbery.util import retry, PlumberyParameters


class RetryableError(Exception):
    pass


class AnotherRetryableError(Exception):
    pass


class UnexpectedError(Exception):
    pass


class TestPlumberyUtil(unittest.TestCase):

    def test_no_retry_required(self):
        self.counter = 0

        @retry(RetryableError, tries=4, delay=0.1)
        def succeeds():
            self.counter += 1
            return 'success'

        r = succeeds()

        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 1)

    def test_retries_once(self):
        self.counter = 0

        @retry(RetryableError, tries=4, delay=0.1)
        def fails_once():
            self.counter += 1
            if self.counter < 2:
                raise RetryableError('failed')
            else:
                return 'success'

        r = fails_once()
        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 2)

    def test_limit_is_reached(self):
        self.counter = 0

        @retry(RetryableError, tries=4, delay=0.1)
        def always_fails():
            self.counter += 1
            raise RetryableError('failed')

        with self.assertRaises(RetryableError):
            always_fails()
        self.assertEqual(self.counter, 4)

    def test_multiple_exception_types(self):
        self.counter = 0

        @retry((RetryableError, AnotherRetryableError), tries=4, delay=0.1)
        def raise_multiple_exceptions():
            self.counter += 1
            if self.counter == 1:
                raise RetryableError('a retryable error')
            elif self.counter == 2:
                raise AnotherRetryableError('another retryable error')
            else:
                return 'success'

        r = raise_multiple_exceptions()
        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 3)

    def test_unexpected_exception_does_not_retry(self):

        @retry(RetryableError, tries=4, delay=0.1)
        def raise_unexpected_error():
            raise UnexpectedError('unexpected error')

        with self.assertRaises(UnexpectedError):
            raise_unexpected_error()

    def test_using_a_logger(self):
        self.counter = 0

        sh = logging.StreamHandler()
        logger = logging.getLogger(__name__)
        logger.addHandler(sh)

        @retry(RetryableError, tries=4, delay=0.1, logger=logger)
        def fails_once():
            self.counter += 1
            if self.counter < 2:
                raise RetryableError('failed')
            else:
                return 'success'

        fails_once()

    def test_parameters(self):
        params = PlumberyParameters()
        self.assertEqual(params.get('dummy'), None)
        self.assertEqual(params.get('dummy', 'ko'), 'ko')
        params.set('dummy', 1)
        self.assertEqual(params.get('dummy'), 1)
        self.assertEqual(params.get('dummy', 'whatever'), 1)
        params.set('dummy', 2)
        self.assertEqual(params.get('dummy'), 2)
        self.assertEqual(params.get('dummy', 'whatever'), 2)

        params = PlumberyParameters({'dummy': 'ok'})
        self.assertEqual(params.get('dummy'), 'ok')
        self.assertEqual(params.get('dummy', 'ko'), 'ok')
        params.set('dummy', 1)
        self.assertEqual(params.get('dummy'), 1)
        self.assertEqual(params.get('dummy', 'whatever'), 1)
        params.set('dummy', 2)
        self.assertEqual(params.get('dummy'), 2)
        self.assertEqual(params.get('dummy', 'whatever'), 2)
        self.assertEqual(params.get('more_dummy'), None)
        self.assertEqual(params.get('more_dummy', 'ok'), 'ok')
        self.assertEqual(params.get('more_dummy', 'ko'), 'ko')
        params.set('more_dummy', 3)
        self.assertEqual(params.get('more_dummy'), 3)
        self.assertEqual(params.get('more_dummy', 'ok'), 3)
        self.assertEqual(params.get('more_dummy', 'ko'), 3)

        with self.assertRaises(TypeError):
            params2 = PlumberyParameters('should raise an error')
        with self.assertRaises(TypeError):
            params2 = PlumberyParameters(params)

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python

"""
Tests for `plogging` module.

"""

import logging
import unittest

from plumbery.plogging import plogging


class TestPlumberyLogging(unittest.TestCase):

    def test_direct(self):

        class DullHandler(logging.NullHandler):
            level = logging.DEBUG
            def emit(self, record):
                log_entry = self.format(record)

        plogging.addHandler(DullHandler())
        plogging.setLevel(logging.DEBUG)
        self.assertEqual(plogging.getEffectiveLevel(), logging.DEBUG)
        plogging.debug("hello world -- debug")
        plogging.info("hello world -- info")
        plogging.warning("hello world -- warning")
        plogging.error("hello world -- error")
        plogging.critical("hello world -- critical")
        self.assertEqual(plogging.foundErrors(), True)
        plogging.reset()
        self.assertEqual(plogging.foundErrors(), False)

    def test_alien(self):
        logger = logging.getLogger('alien.from.mars')
        logger.setLevel(logging.DEBUG)
        logger.debug("hello mars -- debug")
        logger.info("hello mars -- info")
        logger.warning("hello mars -- warning")
        logger.error("hello mars -- error")
        logger.critical("hello mars -- critical")

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python

"""
Tests for `plumbery` module.
"""

# special construct to allow relative import
#
if __name__ == "__main__" and __package__ is None:
    __package__ = "tests"
from tests import dummy

import base64
import logging
import mock
import os
import unittest
import yaml

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import serialization

import six

if six.PY2:
    b = bytes = ensure_string = str
else:
    def ensure_string(s):
        if isinstance(s, str):
            return s
        elif isinstance(s, bytes):
            return s.decode('utf-8')
        else:
            raise TypeError("Invalid argument %r for ensure_string()" % (s,))

    def b(s):
        if isinstance(s, str):
            return s.encode('utf-8')
        elif isinstance(s, bytes):
            return s
        elif isinstance(s, int):
            return bytes([s])
        else:
            raise TypeError("Invalid argument %r for b()" % (s,))

from libcloud.compute.drivers.dimensiondata import DimensionDataNodeDriver

from plumbery.__main__ import parse_args, main
from plumbery.action import PlumberyAction
from plumbery.engine import PlumberyEngine
from plumbery.plogging import plogging
from plumbery.polisher import PlumberyPolisher
from plumbery import __version__

import requests_mock
from .mock_api import DimensionDataMockHttp
DIMENSIONDATA_PARAMS = ('user', 'password')

myParameters = {

    'locationId': 'EU8',
    'domainName': 'aDifferentDomain',
    'networkName': 'aDifferentNetwork'

    }

myPlan = """
---
safeMode: False

information:
  - hello
  - world

links:
  documentation: "http://www.acme.com/"

defaults:

  domain:
    ipv4: auto

  cloud-config:

    disable_root: false
    ssh_pwauth: true
    ssh_keys:
      rsa_private: |
        {{ pair1.rsa_private }}

      rsa_public: "{{ pair1.ssh.rsa_public }}"

    write_files:

    runcmd:

parameters:

  locationId:
    information:
      - "the target data centre for this deployment"
    type: locations.list
    default: EU6

  domainName:
    information:
      - "the name of the network domain to be deployed"
    type: str
    default: myDC

  networkName:
    information:
      - "the name of the Ethernet VLAN to be deployed"
    type: str
    default: myVLAN

buildPolisher: alien

actions:
  - ansible:
      output: gigafox_ansible.yaml
  - inventory:
      output: gigafox_inventory.yaml
  - prepare:
      key: ~/.ssh/myproject_rsa.pub
      output: gigafox_prepares.yaml

---
# Frankfurt in Europe
locationId: "{{ parameter.locationId }}"
regionId: dd-eu

blueprints:

  - myBlueprint:
      domain:
        name: "{{ parameter.domainName }}"
      ethernet:
        name: "{{ parameter.networkName }}"
        subnet: 10.1.10.0
      nodes:
        - myServer:
"""

myBadPlan1 = """
---
parameters:

  locationId:
    information:
      - "the target data centre for this deployment"
    type: locations.list
    default: EU6

  domainName:
    information:
      - "the name of the network domain to be deployed"
    type: str
    default: myDC

  networkName:
    information:
      - "the name of the Ethernet VLAN to be deployed"
    type: str
    default: myVLAN

  parameterWithoutDefaultValue:
    information:
      - "this definition is partial, and missing a defautl value"
    type: str

---
locationId: "{{ parameter.locationId }}"

blueprints:

  - myBlueprint:
      domain:
        name: "{{ parameter.domainName }}"
      ethernet:
        name: "{{ parameter.networkName }}"
        subnet: 10.1.10.0
      nodes:
        - myServer:
"""

myEuropeanPlan = """
---
safeMode: False

information:
  - hello
  - world

links:
  documentation: "http://www.acme.com/"

defaults:

  domain:
    ipv4: auto

  cloud-config:

    disable_root: false
    ssh_pwauth: true
    ssh_keys:
      rsa_private: |
        {{ pair1.rsa_private }}

      rsa_public: "{{ pair1.ssh.rsa_public }}"

    hostname: "{{ parameter.nodeName }}"

    packages:
      - ntp

    write_files:

      - path: /root/hosts.awk
        content: |
          #!/usr/bin/awk -f
          /^{{ {{ parameter.nodeName }}.private }}/ {next}
          /^{{ {{ parameter.nodeName }}.ipv6 }}/ {next}
          {print}
          END {
           print "{{ {{ parameter.nodeName }}.private }}    {{ parameter.nodeName }}"
           print "{{ {{ parameter.nodeName }}.ipv6 }}    {{ parameter.nodeName }}"
          }

parameters:

  locationId:
    information:
      - "the target data centre for this deployment"
    type: locations.list
    default: EU8

  regionId:
    information:
      - "the target region for this deployment"
    type: regions.list
    default: dd-eu

  domainName:
    information:
      - "the name of the network domain to be deployed"
    type: str
    default: myDC

  networkName:
    information:
      - "the name of the Ethernet VLAN to be deployed"
    type: str
    default: myVLAN

  nodeName:
    information:
      - "the name of the node to be deployed"
    type: str
    default: myServer

---
locationId: "{{ parameter.locationId }}"
regionId: {{ parameter.regionId }}

blueprints:

  - myBlueprint:
      domain:
        name: "{{ parameter.domainName }}"
      ethernet:
        name: "{{ parameter.networkName }}"
        subnet: 10.1.10.0
      nodes:
        - {{ parameter.nodeName }}:
"""

myAmericanBinding = {
    'locationId': 'NA9',
    'regionId': 'dd-na',
    'nodeName': 'toto'
    }


myFacility = {
    'regionId': 'dd-na',
    'locationId': 'NA9',
    'blueprints': [{
        'fake': {
            'domain': {
                'name': 'VDC1',
                'service': 'ADVANCED',
                'description': 'fake'},
            'ethernet': {
                'name': 'vlan1',
                'subnet': '10.0.10.0',
                'description': 'fake'},
            'nodes': [{
                'stackstorm': {
                    'description': 'fake',
                    'appliance': 'RedHat 6 64-bit 4 CPU'
                    }
                }]
            }
        }]
    }


myPrivatePlan = """
---
safeMode: True
apiHost: quasimoto.com
locationId: NA9
blueprints:

  - myBlueprint:
      domain:
        name: myDC
      ethernet:
        accept:
          - NA19::remoteNetwork
      nodes:
        - myServer:
            default: bee
            information:
              - complementary information
            memory: 5
            cloud-config:
              packages:
                - smtp
              runcmd:
                - echo "world"

"""


class FakeLocation:

    id = 'EU7'
    name = 'data centre in Amsterdam'
    country = 'Netherlands'


class FakeAction(PlumberyAction):
    def __init__(self, settings):
        self.count = 3
        self.label = 'fake'

    def begin(self, engine):
        self.count += 100

    def enter(self, facility):
        self.count *= 2

    def process(self, blueprint):
        self.count += 5

    def quit(self):
        self.count -= 2

    def end(self):
        self.count += 1


class TestPlumberyEngine(unittest.TestCase):

    def test_init(self):

        engine = PlumberyEngine()

        engine.set_fittings(myPlan)

        self.assertEqual(engine.buildPolisher, 'alien')

        domain = engine.get_default('domain')
        self.assertEqual(domain['ipv4'], 'auto')

        cloudConfig = engine.get_default('cloud-config', {})
        self.assertEqual(len(cloudConfig.keys()), 5)

        self.assertEqual(len(engine.information), 2)

        self.assertEqual(len(engine.links), 1)

        parameters = engine.get_parameters()
        self.assertEqual(parameters['parameter.locationId'],
                         'EU6')
        self.assertEqual(parameters['parameter.domainName'],
                         'myDC')
        self.assertEqual(parameters['parameter.networkName'],
                         'myVLAN')

        parameter = engine.get_parameter('locationId')
        self.assertEqual(parameter, 'EU6')

        parameter = engine.get_parameter('domainName')
        self.assertEqual(parameter, 'myDC')

        parameter = engine.get_parameter('networkName')
        self.assertEqual(parameter, 'myVLAN')

        self.assertEqual(len(engine.polishers), 3)
        for polisher in engine.polishers:
            self.assertTrue(isinstance(polisher, PlumberyPolisher))

        self.assertEqual(engine.safeMode, False)

        self.assertEqual(len(engine.facilities), 1)
        facility = engine.facilities[0]
        self.assertEqual(facility.settings['locationId'], 'EU6')
        self.assertEqual(facility.settings['regionId'], 'dd-eu')
        blueprint = facility.blueprints[0]['myBlueprint']
        self.assertEqual(blueprint['domain']['name'], 'myDC')
        self.assertEqual(blueprint['ethernet']['name'], 'myVLAN')

    def test_parameters(self):

        engine = PlumberyEngine()

        parameters = engine.get_parameters()
        self.assertTrue('parameter.locationId' not in parameters)
        self.assertTrue('parameter.domainName' not in parameters)
        self.assertTrue('parameter.networkName' not in parameters)

        with self.assertRaises(KeyError):
            engine.get_parameter('locationId')
        with self.assertRaises(KeyError):
            engine.get_parameter('domainName')
        with self.assertRaises(KeyError):
            engine.get_parameter('perfectlyUnknownParameter')

        with self.assertRaises(KeyError):
            engine.lookup('parameter.locationId')

        with self.assertRaises(ValueError):
            engine.set_fittings(myBadPlan1)

        engine.set_fittings(myPlan)

        parameters = engine.get_parameters()
        self.assertEqual(parameters['parameter.locationId'],
                         'EU6')
        self.assertEqual(parameters['parameter.domainName'],
                         'myDC')
        self.assertEqual(parameters['parameter.networkName'],
                         'myVLAN')

        self.assertEqual(engine.get_parameter('locationId'),
                         'EU6')
        self.assertEqual(engine.get_parameter('parameter.locationId'),
                         'EU6')
        with self.assertRaises(KeyError):
            engine.get_parameter('perfectlyUnknownParameter')

        engine = PlumberyEngine()

        engine.set_parameters(myParameters)

        parameters = engine.get_parameters()
        self.assertEqual(parameters['parameter.locationId'],
                         'EU8')
        self.assertEqual(parameters['parameter.domainName'],
                         'aDifferentDomain')
        self.assertEqual(parameters['parameter.networkName'],
                         'aDifferentNetwork')

        engine.set_fittings(myPlan)

        parameters = engine.get_parameters()
        self.assertEqual(parameters['parameter.locationId'],
                         'EU8')
        self.assertEqual(parameters['parameter.domainName'],
                         'aDifferentDomain')
        self.assertEqual(parameters['parameter.networkName'],
                         'aDifferentNetwork')

        self.assertEqual(engine.safeMode, False)

        self.assertEqual(len(engine.information), 2)

        self.assertEqual(len(engine.links), 1)

        domain = engine.get_default('domain')
        self.assertEqual(domain['ipv4'], 'auto')

        cloudConfig = engine.get_default('cloud-config', {})
        self.assertEqual(len(cloudConfig.keys()), 5)

        parameter = engine.get_parameter('locationId')
        self.assertEqual(parameter, 'EU8')

        parameter = engine.get_parameter('domainName')
        self.assertEqual(parameter, 'aDifferentDomain')

        parameter = engine.get_parameter('networkName')
        self.assertEqual(parameter, 'aDifferentNetwork')

        self.assertEqual(len(engine.facilities), 1)
        facility = engine.facilities[0]
        self.assertEqual(facility.settings['locationId'], 'EU8')
        self.assertEqual(facility.settings['regionId'], 'dd-eu')
        blueprint = facility.blueprints[0]['myBlueprint']
        self.assertEqual(blueprint['domain']['name'], 'aDifferentDomain')
        self.assertEqual(blueprint['ethernet']['name'], 'aDifferentNetwork')

    def test_environment(self):

        engine = PlumberyEngine()
        self.assertTrue(len(engine.lookup('environment.PATH')) > 0)
        with self.assertRaises(KeyError):
            engine.lookup('environment.PERFECTLY_UNKNOWN_FROM_HERE')

    def test_set(self):

        engine = PlumberyEngine()
        DimensionDataNodeDriver.connectionCls.conn_classes = (
            None, DimensionDataMockHttp)
        DimensionDataMockHttp.type = None
        self.region = DimensionDataNodeDriver(*DIMENSIONDATA_PARAMS)

        file = os.path.abspath(
            os.path.dirname(__file__))+'/fixtures/dummy_rsa.pub'

        settings = {
            'keys': [ "*hello-there*" ],
            }

        with self.assertRaises(ValueError):
            engine.set_settings(settings)

        settings = {
            'keys': [ file ],
            }

        engine.set_settings(settings)
        self.assertTrue(isinstance(engine.get_shared_key_files(), list))
        self.assertTrue(file in engine.get_shared_key_files())

        settings = {
            'safeMode': False,
            'polishers': [
                {'ansible': {}},
                {'configure': {}},
                ],
            'keys': [ file, file ],
            }

        engine.set_settings(settings)
        self.assertEqual(engine.safeMode, False)
        self.assertTrue(isinstance(engine.get_shared_key_files(), list))
        self.assertTrue(file in engine.get_shared_key_files())

        engine.add_facility(myFacility)
        self.assertEqual(len(engine.facilities), 1)

        self.assertEqual(engine.get_shared_user(), 'root')
        engine.set_shared_user('ubuntu')
        self.assertEqual(engine.get_shared_user(), 'ubuntu')

        engine.set_shared_secret('fake_secret')
        self.assertEqual(engine.get_shared_secret(), 'fake_secret')

        random = engine.get_secret('random')
        self.assertEqual(len(random), 9)
        self.assertEqual(engine.get_secret('random'), random)

        engine.set_user_name('fake_name')
        self.assertEqual(engine.get_user_name(), 'fake_name')

        engine.set_user_password('fake_password')
        self.assertEqual(engine.get_user_password(), 'fake_password')

    def test_settings_private(self):
        engine = PlumberyEngine()
        engine.set_shared_secret('fake_secret')
        engine.set_user_name('fake_name')
        engine.set_user_password('fake_password')
        engine.set_fittings(myPrivatePlan)
        facilities = engine.list_facility('quasimoto.com')
        self.assertEqual(len(facilities), 1)
        facilities[0].power_on()
#        self.assertEqual(facilities[0].region.connection.host, 'quasimoto.com')

    def test_lifecycle(self):

        engine = PlumberyEngine()
        DimensionDataNodeDriver.connectionCls.conn_classes = (
            None, DimensionDataMockHttp)
        DimensionDataMockHttp.type = None
        self.region = DimensionDataNodeDriver(*DIMENSIONDATA_PARAMS)

        engine.set_shared_secret('fake_secret')
        engine.set_user_name('fake_name')
        engine.set_user_password('fake_password')

        engine.do('build')
        engine.build_all_blueprints()
        engine.do('build', 'myBlueprint')
        engine.build_blueprint('myBlueprint')

        engine.do('deploy')
        engine.do('deploy', 'myBlueprint')

        engine.do('destroy')
        engine.destroy_all_blueprints()
        engine.do('destroy', 'myBlueprint')
        engine.destroy_blueprint('myBlueprint')

        engine.do('dispose')
        engine.do('dispose', 'myBlueprint')

        engine.do('polish')
        engine.polish_all_blueprints()
        engine.do('polish', 'myBlueprint')
        engine.polish_blueprint('myBlueprint')

        engine.do('secrets')

        engine.do('start')
        engine.start_all_blueprints()
        engine.do('start', 'myBlueprint')
        engine.start_blueprint('myBlueprint')

        engine.do('stop')
        engine.stop_all_blueprints()
        engine.do('stop', 'myBlueprint')
        engine.stop_blueprint('myBlueprint')

        engine.do('wipe')
        engine.wipe_all_blueprints()
        engine.do('wipe', 'myBlueprint')
        engine.wipe_blueprint('myBlueprint')

        banner = engine.document_elapsed()
        self.assertEqual('Worked for you' in banner, True)

    def test_process_all_blueprints(self):

        engine = PlumberyEngine()
        DimensionDataNodeDriver.connectionCls.conn_classes = (
            None, DimensionDataMockHttp)
        DimensionDataMockHttp.type = None
        self.region = DimensionDataNodeDriver(*DIMENSIONDATA_PARAMS)

        engine.set_shared_secret('fake_secret')
        engine.set_user_name('fake_name')
        engine.set_user_password('fake_password')
        engine.set_fittings(myPrivatePlan)

        engine.process_all_blueprints(action='noop')

        action = FakeAction({})
        engine.process_all_blueprints(action)
        self.assertEqual(action.count, 210)

    def test_process_blueprint(self):

        engine = PlumberyEngine()
        DimensionDataNodeDriver.connectionCls.conn_classes = (
            None, DimensionDataMockHttp)
        DimensionDataMockHttp.type = None
        self.region = DimensionDataNodeDriver(*DIMENSIONDATA_PARAMS)

        engine.set_shared_secret('fake_secret')
        engine.set_user_name('fake_name')
        engine.set_user_password('fake_password')
        engine.set_fittings(myPrivatePlan)

        engine.process_blueprint(action='noop', names='fake')

        action = FakeAction({})
        engine.process_blueprint(action, names='fake')
        self.assertEqual(action.count, 205)

    def test_as_library(self):

        engine = PlumberyEngine(myEuropeanPlan, myAmericanBinding)
        DimensionDataNodeDriver.connectionCls.conn_classes = (
            None, DimensionDataMockHttp)
        DimensionDataMockHttp.type = None
        self.region = DimensionDataNodeDriver(*DIMENSIONDATA_PARAMS)

        engine.set_shared_secret('fake_secret')
        engine.set_user_name('fake_name')
        engine.set_user_password('fake_password')

        facilities = engine.list_facility('NA9')
        self.assertEqual(len(facilities), 1)

        facility = facilities[0]
        self.assertEqual(facility.get_setting('regionId'), 'dd-na')
        self.assertEqual(facility.get_setting('locationId'), 'NA9')

        self.assertTrue(facility.get_blueprint('fake') is None)

        blueprint = facility.get_blueprint('myBlueprint')

        node = blueprint['nodes'][0]
        self.assertEqual(list(node)[0], 'toto')

        config = node['toto']['cloud-config']
        self.assertEqual(config['hostname'], 'toto')
        self.assertEqual(config['write_files'][0]['content'].count('toto'), 6)

        engine.do('deploy')
        engine.do('dispose')

    def test_lookup(self):

        engine = PlumberyEngine()
        self.assertEqual(engine.lookup('plumbery.version'), __version__)

        engine.secrets = {}
        random = engine.lookup('secret.random')
        self.assertEqual(len(random), 9)
        self.assertEqual(engine.lookup('secret.random'), random)

        md5 = engine.lookup('secret.random.md5')
        self.assertEqual(len(md5), 32)
        self.assertNotEqual(md5, random)

        sha = engine.lookup('secret.random.sha1')
        self.assertEqual(len(sha), 40)
        self.assertNotEqual(sha, random)

        sha = engine.lookup('secret.random.sha256')
        self.assertEqual(len(sha), 64)
        self.assertNotEqual(sha, random)

        id1 = engine.lookup('id1.uuid')
        self.assertEqual(len(id1), 36)
        self.assertEqual(engine.lookup('id1.uuid'), id1)
        id2 = engine.lookup('id2.uuid')
        self.assertEqual(len(id2), 36)
        self.assertNotEqual(id1, id2)

        engine.lookup('application.secret')
        engine.lookup('database.secret')
        engine.lookup('master.secret')
        engine.lookup('slave.secret')

        original = b'hello world'
        print('original: {}'.format(original))

        text = ensure_string(engine.lookup('rsa_public.pair1'))
        print('rsa_public.pair1: {}'.format(text))

        self.assertTrue(text.startswith('ssh-rsa '))

        text = b(text)
        key = serialization.load_ssh_public_key(
            data=text,
            backend=default_backend())

        encrypted = key.encrypt(
            original,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA1()),
                algorithm=hashes.SHA1(),
                label=None
            )
        )
        encrypted = base64.b64encode(encrypted)
        print('encrypted: {}'.format(encrypted))

        privateKey = engine.lookup('rsa_private.pair1')
        print('rsa_private.pair1: {}'.format(privateKey))
        self.assertTrue(ensure_string(privateKey).startswith(
            '-----BEGIN RSA PRIVATE KEY-----'))

        privateKey = serialization.load_pem_private_key(
            b(privateKey),
            password=None,
            backend=default_backend())

        decrypted = privateKey.decrypt(
            base64.b64decode(encrypted),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA1()),
                algorithm=hashes.SHA1(),
                label=None
            )
        )
        print('decrypted: {}'.format(decrypted))

        self.assertEqual(decrypted, original)

        token = engine.lookup('https://discovery.etcd.io/new')
        self.assertEqual(token.startswith(
            'https://discovery.etcd.io/'), True)
        self.assertEqual(len(token), 58)

        self.assertEqual(len(engine.secrets), 13)

        with self.assertRaises(LookupError):
            localKey = engine.lookup('rsa_private.local')

        localKey = engine.lookup('rsa_public.local')
        if len(localKey) > 0:
            path = engine.get_shared_key_files()[0]
            with open(os.path.expanduser(path)) as stream:
                text = stream.read()
                stream.close()
                self.assertEqual(localKey.strip(), text.strip())
                plogging.info("Successful lookup of local public key")

    def test_secrets(self):

        engine = PlumberyEngine()
        engine.secrets = {'hello': 'world'}
        engine.save_secrets(plan='test_engine.yaml')
        engine.secrets = {}
        engine.load_secrets(plan='test_engine.yaml')
        self.assertEqual(engine.secrets['hello'], 'world')
        engine.forget_secrets(plan='test_engine.yaml')
        self.assertEqual(os.path.isfile('.test_engine.secrets'), False)

    def test_keys(self):

        engine = PlumberyEngine()
        self.assertEqual(engine._sharedKeyFiles, [])

        with self.assertRaises(ValueError):
            engine.set_shared_key_files('this_does_not_exist')
        self.assertTrue(isinstance(engine.get_shared_key_files(), list))
        self.assertEqual(engine._sharedKeyFiles,
                         engine.get_shared_key_files())

        with mock.patch.object(engine, 'get_shared_key_files') as patched:
            patched.return_value = []

            files = ['*unknown-file*', '**me-too*']
            with self.assertRaises(ValueError):
                engine.set_shared_key_files(files)

        file = os.path.abspath(
            os.path.dirname(__file__))+'/fixtures/dummy_rsa.pub'
        engine.set_shared_key_files(file)
        self.assertTrue(isinstance(engine.get_shared_key_files(), list))
        self.assertEqual(engine.get_shared_key_files()[0], file)
        self.assertEqual(engine._sharedKeyFiles,
                         engine.get_shared_key_files())

        with mock.patch.object(engine, 'get_shared_key_files') as patched:
            patched.return_value = []

            files = [file, file, file]
            engine.set_shared_key_files(files)

            self.assertEqual(engine._sharedKeyFiles, [file])

        if 'SHARED_KEY' in os.environ:
            memory = os.environ["SHARED_KEY"]
        else:
            memory = None

        engine._sharedKeyFiles = []
        os.environ["SHARED_KEY"] = 'this_does_not_exist'
        with self.assertRaises(ValueError):
            engine.set_shared_key_files()
        with self.assertRaises(ValueError):
            engine.get_shared_key_files()
        self.assertTrue(isinstance(engine._sharedKeyFiles, list))

        engine._sharedKeyFiles = []
        os.environ["SHARED_KEY"] = file
        engine.set_shared_key_files()
        self.assertTrue(isinstance(engine.get_shared_key_files(), list))
        self.assertEqual(engine.get_shared_key_files()[0], file)
        self.assertEqual(engine._sharedKeyFiles,
                         engine.get_shared_key_files())

        if memory is None:
            os.environ.pop("SHARED_KEY")
        else:
            os.environ["SHARED_KEY"] = memory

        with mock.patch.object(engine, 'get_shared_key_files') as patched:
            patched.return_value = []

            self.assertTrue(isinstance(engine.get_shared_key_files(), list))

            engine.set_shared_key_files()
            self.assertTrue(plogging.foundErrors())

            with self.assertRaises(ValueError):
                engine.set_shared_key_files('this_does_not_exist')

            file = os.path.abspath(
                os.path.dirname(__file__))+'/fixtures/dummy_rsa.pub'
            engine.set_shared_key_files(file)
            self.assertEqual(engine._sharedKeyFiles, [file])

    def test_parser(self):

        args = parse_args(['fittings.yaml', 'build', 'web'])
        self.assertEqual(args.fittings, 'fittings.yaml')
        self.assertEqual(args.action, 'build')
        self.assertEqual(args.blueprints, ['web'])
        self.assertEqual(args.facilities, None)

        args = parse_args(
            ['fittings.yaml', 'build', 'web', '-p', 'parameters.yaml'])
        self.assertEqual(args.parameters, ['parameters.yaml'])

        args = parse_args(
            ['fittings.yaml', 'build', 'web', '-p', 'parameters.yaml', '-s'])
        self.assertEqual(args.parameters, ['parameters.yaml'])
        self.assertEqual(args.safe, True)

        args = parse_args(
            ['fittings.yaml', 'build', 'web', '-p', 'parameters.yaml', '-d'])
        self.assertEqual(args.parameters, ['parameters.yaml'])
        self.assertEqual(args.debug, True)

        args = parse_args(['fittings.yaml', 'build', 'web', '-s'])
        self.assertEqual(args.safe, True)

        args = parse_args(['fittings.yaml', 'build', 'web', '-d'])
        self.assertEqual(args.debug, True)
        self.assertEqual(
            plogging.getEffectiveLevel(), logging.DEBUG)

        args = parse_args(['fittings.yaml', 'build', 'web', '-q'])
        self.assertEqual(args.quiet, True)
        self.assertEqual(
            plogging.getEffectiveLevel(), logging.WARNING)

        args = parse_args(['fittings.yaml', 'start', '@NA12'])
        self.assertEqual(args.fittings, 'fittings.yaml')
        self.assertEqual(args.action, 'start')
        self.assertEqual(args.blueprints, None)
        self.assertEqual(args.facilities, ['NA12'])

        args = parse_args([
            'fittings.yaml', 'prepare', 'web', 'sql', '@NA9', '@NA12'])
        self.assertEqual(args.fittings, 'fittings.yaml')
        self.assertEqual(args.action, 'prepare')
        self.assertEqual(args.blueprints, ['web', 'sql'])
        self.assertEqual(args.facilities, ['NA9', 'NA12'])

        args = parse_args([
            'fittings.yaml', 'prepare', 'web', '@NA9', 'sql', '@NA12'])
        self.assertEqual(args.fittings, 'fittings.yaml')
        self.assertEqual(args.action, 'prepare')
        self.assertEqual(args.blueprints, ['web', 'sql'])
        self.assertEqual(args.facilities, ['NA9', 'NA12'])

        args = parse_args(['fittings.yaml', 'polish'])
        self.assertEqual(args.fittings, 'fittings.yaml')
        self.assertEqual(args.action, 'polish')
        self.assertEqual(args.blueprints, None)
        self.assertEqual(args.facilities, None)

    def test_main(self):

        with self.assertRaises(SystemExit):
            main(['fittings.yaml', 'build', 'web', '@EU6'])

        engine = PlumberyEngine()
        engine.set_fittings(myPlan)
        engine.set_user_name('fake_name')
        engine.set_user_password('fake_password')
        with self.assertRaises(SystemExit):
            main(['-v'], engine)
        with self.assertRaises(SystemExit):
            main(['fittings.yaml', 'build', 'web'], engine)
        with self.assertRaises(SystemExit):
            main(['fittings.yaml', 'build', 'web', '-v'], engine)
        with self.assertRaises(SystemExit):
            main(['fittings.yaml', 'build', 'web', '@EU6'], engine)

    def test_bad_args(self):

        engine = PlumberyEngine()
        engine.set_fittings(myPlan)
        with self.assertRaises(SystemExit):
            main(['bad args'], engine)
        with self.assertRaises(SystemExit):
            main(['fittings.yaml'], engine)
        with self.assertRaises(SystemExit):
            main(['fittings.yaml', 'xyz123', 'web'], engine)
        with self.assertRaises(SystemExit):
            main(['fittings.yaml', 'build', 'web', '@'], engine)

    def test_param_http(self):
        engine = PlumberyEngine()
        with self.assertRaises(TypeError):
            engine.set_parameters(('http://smee.com/params.yml'))

    def test_remote_params(self):
        engine = PlumberyEngine()
        with requests_mock.mock() as m:
            m.get('http://smee.com/params.yml', text=yaml.dump(myParameters))
            engine.set_parameters('http://smee.com/params.yml')

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import time
from functools import wraps

from plumbery.plogging import plogging


def retry(ExceptionToCheck, tries=4, delay=3, backoff=2, logger=plogging):
    """
    Retries calling the decorated function using an exponential backoff.

    http://thecodeship.com/patterns/guide-to-python-function-decorators/

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :type ExceptionToCheck: Exception or tuple

    :param tries: number of times to try (not retry) before giving up
    :type tries: int

    :param delay: initial delay between retries in seconds
    :type delay: int

    :param backoff: backoff multiplier e.g. value of 2 will double the delay
        each retry
    :type backoff: int

    :param logger: logger to use. If None, print
    :type logger: logging.Logger instance

    """
    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    msg1 = "{}".format(str(e))
                    msg2 = "- retrying in {} seconds...".format(mdelay)
                    if logger:
                        logger.warning(msg1)
                        logger.warning(msg2)
                    else:
                        print(msg1)
                        print(msg2)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry


class PlumberyParameters(object):
    """
    Manages parameters

    :param dictionary: pre-populated set of named parameters
    :type dictionary: ``dict``

    """

    def __init__(self, dictionary=None):

        if dictionary is None:
            self.dictionary = {}
        elif not isinstance(dictionary, dict):
            raise TypeError("invalid type for parameter 'dictionary'")
        else:
            self.dictionary = dictionary

    def set(self, parameter, value):
        """
        Remembers the value of a parameter

        :param parameter: the parameter
        :type parameter: ``str``

        :param value: the value
        :type value: any

        """

        self.dictionary[parameter] = value

    def get(self, parameter, default=None):
        """
        Retrieves the value of a parameter

        :param parameter: the parameter
        :type parameter: ``str``

        :param default: default value to return if the parameter has not been set
        :type default: any

        :return: the value of this parameter, or None

        """

        if parameter in self.dictionary:
            return self.dictionary[parameter]

        return default



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from plumbery.plogging import plogging
from plumbery.util import PlumberyParameters

__all__ = ['PlumberyAction', 'PlumberyActionLoader']


class PlumberyAction(object):
    """
    Performs some action on elements described in the fittings plan

    :param parameters: specific parameters for this action
    :type parameters: :class:`PlumberyParameters` or ``dict``

    An action is an imperative given to plumbery, such as
    'build', 'stop' or 'dispose'. It is commonly expressed as a verb, but
    not always. For example, 'information' asks plumbery to describe
    elements of the fittings plan.

    The engine provides a simple way to trigger an action. For example::

        from plumbery.engine import PlumberyEngine
        PlumberyEngine('fittings.yaml').do('start')

    """

    def __init__(self, parameters=PlumberyParameters()):
        if parameters is None:
            parameters = PlumberyParameters()
        elif not isinstance(parameters, PlumberyParameters):
            raise TypeError("'parameters' should have type PlumberyParameters")

        self.parameters = parameters

    def get_banner(self, location):
        """
        Informs the end-user
        """

        return "Plumbing at {}".format(location)

    def get_label(self):
        """
        Provides the label of the action

        :return: The label, such as 'build' or 'stop'
        :rtype: ``str``

        """

        return self.label

    def get_parameter(self, name, default=None):
        """
        Provides settings for this action

        :param name: the parameter
        :type name: ``str``

        :param default: returned if the parameter has not been set
        :type default: any

        :return: the value of this parameter, or None

        """

        return self.parameters.get(name, default)

    def begin(self, engine):
        """
        Binds to a plumbery engine

        :param engine: the plumbery engine itself
        :type engine: :class:`plumbery.PlumberyEngine`

        This function is called once, before all others.
        You can override it for any specific initialisation.

        """

        self.engine = engine

        plogging.info("Using '{}'".format(self.get_label()))

    def enter(self, facility):
        """
        Enters a facility to process blueprints there

        :param facility: a targeted facility
        :type facility: :class:`plumbery.PlumberyFacility`

        This function is called once for each facility that is visited during
        the process. You can override it for any specific
        initialisation that you would require.

        """

        self.facility = facility

        plogging.info(self.get_banner(facility))

    def process(self, blueprint):
        """
        Prepares one blueprint

        :param blueprint: a targeted blueprint
        :type blueprint: :class:`plumbery.PlumberyBlueprint`

        This is where the hard work is done. You have to override this
        function in your own action.

        """

        pass

    def quit(self):
        """"
        Quits a facility

        This function can be useful for some cleaning after processing.
        """

        pass

    def end(self):
        """
        Reaps the outcome of all this action

        This function is called once, at the very end of the process. You
        can override it for any specific closure that you would require.

        """

        pass


class PlumberyActionLoader(object):
    """
    Loads action classes dynamically

    Plumbery is coming with a number of pre-defined actions that, together,
    cover the entire life cycle of virtual assets in the cloud.

    You can create actions of your own, or use actions from other
    persons. All actions have to be placed in the directory
    ``plumbery/actions``. Each action should be put in a separate
    python file, and define a class that repeats the file name. For example
    the file::

        plumbery/actions/configure.py

    should contain::

        class ConfigureAction(PlumberyAction):
        ...

    Once this is done properly, you can use the action from the command line,
    or via the API.

    Also, you can pass any parameters that the action would require.

    Example of configuration of ``fittings.yaml``::

        ---
        safeMode: False
        actions:
          - configure:
              file: nodes.yaml
          - ansible:
              file: inventory.yaml
        ---
        # Frankfurt in Europe
        locationId: EU6
        regionId: dd-eu
        ...

    """

    @classmethod
    def load_all(cls, settings={}):

        actions = {}

        there = os.path.abspath(os.path.dirname(__file__))+'/actions'
        for unused1, unused2, files in os.walk(there):
            plogging.debug("Loading actions from %s", there)

            for file in files:
                label, extension = file.split('.', 1)
                if len(label) < 2:
                    continue
                if label[0] in ('-', '_', '~'):
                    continue
                if extension == 'py':
                    plogging.debug("Loading action %s", file)

                    if label in settings:
                        parameters = settings[ label ]
                    else:
                        parameters = {}

                    actions[ label ] = PlumberyActionLoader.load(label,
                                                                 parameters)

        return actions


    @classmethod
    def load(cls, label, parameters=PlumberyParameters()):
        """
        Picks up an action from the shelf

        :param label: name of the action to use, e.g., ``inventory``
        :type label: ``str``

        :param parameters: specific parameters for this action
        :type parameters: ``dict``

        :return: instance of a action ready to use
        :rtype: :class:`plumbery.PlumberyAction`

        """

        moduleName = 'plumbery.actions.' + label
        actionName = label.capitalize() + 'Action'
        try:

            plogging.debug("Importing '{}'".format(moduleName))
            actionModule = __import__(
                moduleName,
                globals(),
                locals(),
                [actionName])

            plogging.debug("Instantiating '{}'".format(actionName))
            actionClass = getattr(actionModule, actionName)
            if parameters is None:
                parameters = PlumberyParameters()
            elif isinstance(parameters, dict):
                parameters = PlumberyParameters(parameters)
            action = actionClass(parameters)
            action.label = label
            return action

        except ImportError as feedback:
            plogging.debug("Unable to find module '{}'".format(moduleName))
            raise

        except TypeError as feedback:
            plogging.debug("Invalid parameters for '{}'".format(moduleName))
            raise

        except ValueError as feedback:
            plogging.debug("Invalid parameters for '{}'".format(moduleName))
            raise

        except Exception as feedback:
            plogging.debug("Unable to import '{}' from '{}'".format(
                actionName, moduleName))
            raise



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess

from plumbery.plogging import plogging


class Terraform(object):
    def __init__(self, working_directory):
        self.working_directory = working_directory
        self.tf_path = os.getenv('TERRAFORM_PATH')
        if self.tf_path is None or not os.path.exists(self.tf_path):
            plogging.debug("Could not locate terraform binary. "
                           "Please check TERRAFORM_PATH ENV var."
                           "Ignore if no multicloud fittings are present")

    def build(self, settings):
        tf_path = settings.get('tf_path', None)
        if tf_path is None:
            # default back to the directory of the fittings file.
            tf_path = self.working_directory

        parameters = settings.get('parameters', {})
        with open(os.path.join(tf_path, '.tfvars'), 'w') as tf_vars:
            for (key, value) in parameters.items():
                tf_vars.write('%s = "%s"\n' % (key, value))

        ret, o, err = self._run_tf(
            'plan', tf_path,
            var_file=os.path.join(tf_path, '.tfvars'),
            input=False,
            detailed_exitcode=True,
            out=os.path.join(tf_path, '.tfstate'))
        plogging.debug("STDOUT from terraform plan %s", o)
        if err != '' or None:
            plogging.error(err)

        if ret == 2:
            _, o, err = self._run_tf('apply', os.path.join(tf_path, '.tfstate'))
            plogging.debug("STDOUT from terraform apply %s", o)
            if err != '' or None:
                plogging.error(err)
        if os.path.isfile(os.path.join(tf_path, '.tfstate')):
            os.remove(os.path.join(tf_path, '.tfstate'))
        if os.path.isfile(os.path.join(tf_path, '.tfvars')):
            os.remove(os.path.join(tf_path, '.tfvars'))

    def destroy(self, settings, safe=True):
        tf_path = settings.get('tf_path', None)
        if tf_path is None:
            # default back to the directory of the fittings file.
            tf_path = self.working_directory

        parameters = settings.get('parameters', {})
        with open(os.path.join(tf_path, '.tfvars'), 'w') as tf_vars:
            for (key, value) in parameters.items():
                tf_vars.write('%s = "%s"\n' % (key, value))
        if safe:
            _, o, _ = self._run_tf(
                'plan', tf_path,
                var_file=os.path.join(tf_path, '.tfvars'),
                input=False,
                detailed_exitcode=True,
                destroy=True)
            plogging.debug("STDOUT from terraform %s", o)
        else:
            _, o, _ = self._run_tf(
                'destroy', tf_path,
                var_file=os.path.join(tf_path, '.tfvars'),
                input=False,
                force=True)
            plogging.debug("STDOUT from terraform %s", o)

    def graph(self, state_directory):
        _, graph_data, _ = self._run_tf('graph', state_directory)
        graph_file_path = os.path.join(state_directory, 'multicloud.dot')
        with open(graph_file_path, 'w') as dot:
            dot.write(graph_data)

    def _run_tf(self, command, state_directory, **kwargs):
        if self.tf_path is None:
            plogging.error("Could not locate terraform binary. "
                           "Please check TERRAFORM_PATH ENV var.")
            raise RuntimeError("Missing terraform binary")
        params = [self.tf_path, command]
        for (key, value) in kwargs.items():
            params.append("-%s=%s" % (key.replace('_', '-'), value))
        params.append(state_directory)
        plogging.debug(params)
        process = subprocess.Popen(params, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        retcode = process.returncode
        return (retcode, stdout, stderr)



#!/usr/bin/env python

"""
Tests for `bootstrap` module.
"""

import unittest
import os
from tempfile import gettempdir
from mock import MagicMock, patch
import requests_mock

import plumbery.bootstrap as b

tempdir = gettempdir()


class BootstrapTests(unittest.TestCase):
    def test_url_cwd(self):
        url = 'http://test.com/fittings.yaml'
        args = b.parse_args([url])
        b.download_file = MagicMock(return_value=True)
        b.main(args)
        b.download_file.assert_called_with(url, os.getcwd())

    def test_url_target_directory(self):
        url = 'http://test.com/fittings.yaml'
        args = b.parse_args([url, '-o', '/tmp'])
        b.download_file = MagicMock(return_value=True)
        b.main(args)
        b.download_file.assert_called_with(url, '/tmp')

    def test_url_manifest(self):
        url = 'http://test.com/manifest.mf'
        args = b.parse_args([url, '-o', '/tmp'])
        b.download_manifest = MagicMock(return_value=True)
        b.main(args)
        b.download_manifest.assert_called_with(url, '/tmp')

    def test_download_manifest(self):
        url = 'http://test.com/manifest.mf'
        b.download_file = MagicMock(return_value='test.file')
        open_name = 'plumbery.bootstrap.open'
        with patch(open_name, create=True) as mock_open:
            mock_open.return_value = MagicMock(return_value='test.file')
            b.download_manifest(url, os.getcwd())
        b.download_file.assert_called_with(url, os.getcwd())

    def test_download_file(self):
        with requests_mock.mock() as m:
            m.get('http://test.com/manifest.mf', text='testing')
            b.download_file('http://test.com/manifest.mf', tempdir)
            with open(os.path.join(tempdir, 'manifest.mf')) as tmp_file:
                self.assertEqual(tmp_file.read(), 'testing')

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())
